{ Code Generator
 (C) 2005-2012, Andreas Berg. All Rights Reserved. }
unit _codegen;

interface

uses
 _common;

const
 codeNONE          = 0;
 codeVALUE         = 1;
 codeREL           = 2;
 codeRELID         = 3;
 codeMEM           = 4;
 codeMEMID         = 5;
 codeEQUOFFSET     = 6;
 codeEQUSIZE       = 7;
 codeGETSEGOFFSET  = 64;
 codeGETSEGSIZE    = 65;
 codeLOCALMODE     = 80;
 codeGLOBALMODE    = 81;
 codeEND           = 254;
 codeRESETSEG      = 255;

type
 TCodeList = record
    SegIndex       : TDefValue;
    ID             : TDefValue;
    IdentName      : string;
    Index          : TDefValue;
    Type1          : byte;
    Type2          : byte;
    Value1         : TAsmValue;
    Value2         : TAsmValue;
    Opr            : byte;
    StringOfs      : TDefValue;
    StringLen      : TAsmValue;
 end;

var
 CodeList          : array of TCodeList;
 CodeListItems     : TAsmValue;

procedure InitCodeLists;
procedure UpdateCodeID;

procedure AddCodeItem(IDV: TDefValue; IName: string;
 Indx: TDefValue; Typ1, Typ2: byte; Val1, Val2: TAsmValue;
 opr1: byte; StrOfs: TDefValue; StrLen: TAsmValue);

procedure AddCodeListLOCAL;
procedure AddCodeListGLOBAL;

procedure AddCodeListValue(vTyp: byte; val: TAsmValue);
procedure AddCodeListREL(vTyp: byte; n: string; o: byte; val: TAsmValue);
procedure AddCodeListMEM(vTyp: byte; n: string; o: byte; val: TAsmValue);
procedure AddCodeListEQUOFFSET(vTyp: byte; n: string; o: byte; val: TAsmValue);
procedure AddCodeListEQUSIZE(vTyp: byte; n: string; o: byte; val: TAsmValue);
procedure AddCodeListSEGOFFSET(vTyp: byte; segIndex: TDefValue;
 o: byte; val: TAsmValue);
procedure AddCodeListSEGSIZE(vTyp: byte; segIndex: TDefValue;
 o: byte; val: TAsmValue);
procedure AddCodeListRELID(vTyp: byte; id: TDefValue;
 o: byte; val: TAsmValue);
procedure AddCodeListMEMID(vTyp: byte; id: TDefValue;
 o: byte; val: TAsmValue);

implementation

uses
 _error, _ident, _seg, _scanner, _parser;

procedure InitCodeLists;
begin
   SetLength(CodeList, 1);
   CodeListItems:= 0;
end;

procedure UpdateCodeID;
var
 i, indx           : TDefValue;
 o                 : byte;
 v, equV           : TAsmValue;
 iMod              : byte;
begin
   UpdateEquates;
   iMod:= IdentMode;
   IdentMode:= imLOCAL;
   if CodeListItems > 0 then begin
      i:= 0;
      while i < CodeListItems do begin
         if CodeList[i].Index = NotFound then begin
            indx:= GetIdentIndex(CodeList[i].IdentName);
            if indx <> NotFound then begin
               CodeList[i].Index:= indx;
               if CodeList[i].ID = codeREL then CodeList[i].ID:= codeRELID;
               if CodeList[i].ID = codeMEM then CodeList[i].ID:= codeMEMID;
               if CodeList[i].ID = codeEQUOFFSET then begin
                  o:= CodeList[i].Opr;
                  v:= CodeList[i].Value1;
                  equV:= IdentList[indx].Offset;
                  CodeList[i].ID:= codeVALUE;
                  CodeList[i].Value1:= DoCalc(equV, o, v);
               end;
               if CodeList[i].ID = codeEQUSIZE then begin
                  o:= CodeList[i].Opr;
                  v:= CodeList[i].Value1;
                  equV:= IdentList[indx].Len;
                  CodeList[i].ID:= codeVALUE;
                  CodeList[i].Value1:= DoCalc(equV, o, v);
               end;
               if CodeList[i].ID = codeLOCALMODE then IdentMode:= imLOCAL;
               if CodeList[i].ID = codeGLOBALMODE then IdentMode:= imGLOBAL;
            end;
         end;
         Inc(i);
      end;
   end;
   IdentMode:= iMod;
end;



procedure AddCodeItem(IDV: TDefValue; IName: string;
 Indx: TDefValue; Typ1, Typ2: byte; Val1, Val2: TAsmValue;
 opr1: byte; StrOfs: TDefValue; StrLen: TAsmValue);
var
 cSeg              : TDefValue;
 vSize             : byte;
begin
   cSeg:= GetSeg;
   if cSeg <> Neg then begin
      vSize:= 1;
      if Typ1 = tWORD then vSize:= 2;
      if Typ1 = tDWORD then vSize:= 4;
      if Typ1 = tQWORD then vSize:= 8;
      if Typ1 = tFARPTR then vSize:= 6;
      if Typ1 = tTENBYTE then vSize:= 10;
      if Typ1 = 255 then vSize:= 0;
      SetLength(CodeList, CodeListItems + 1);
      CodeList[CodeListItems].SegIndex:= cSeg;
      CodeList[CodeListItems].ID:= IDV;
      CodeList[CodeListItems].IdentName:= IName;
      CodeList[CodeListItems].Index:= Indx;
      CodeList[CodeListItems].Type1:= Typ1;
      CodeList[CodeListItems].Type2:= Typ2;
      CodeList[CodeListItems].Value1:= Val1;
      CodeList[CodeListItems].Value2:= VAl2;
      CodeList[CodeListItems].Opr:= opr1;
      CodeList[CodeListItems].StringOfs:= StrOfs;
      CodeList[CodeListItems].StringLen:= StrLen;
      Inc(CodeListItems);
      AddSegSize(vSize);
   end else begin
      ErrorMsg('Undefined segment [Can not write data]!');
   end;
end;



procedure AddCodeListLOCAL;
begin
   AddCodeItem(codeLOCALMODE, '', NotFound, 255, tNONE,
    0, 0, oprNONE, 0, 0);
end;

procedure AddCodeListGLOBAL;
begin
   AddCodeItem(codeGLOBALMODE, '', NotFound, 255, tNONE,
    0, 0, oprNONE, 0, 0);
end;



procedure AddCodeListValue(vTyp: byte; val: TAsmValue);
begin
   AddCodeItem(codeValue, '', NotFound, vTyp, tNONE,
    val, 0, oprNONE, 0, 0);
end;

procedure AddCodeListREL(vTyp: byte; n: string; o: byte; val: TAsmValue);
begin
   AddCodeItem(codeREL, n, NotFound, vTyp, tNONE,
    val, 0, o, 0, 0);
end;

procedure AddCodeListMEM(vTyp: byte; n: string; o: byte; val: TAsmValue);
begin
   AddCodeItem(codeMEM, n, NotFound, vTyp, tNONE,
    val, 0, o, 0, 0);
end;

procedure AddCodeListEQUOFFSET(vTyp: byte; n: string; o: byte; val: TAsmValue);
begin
   AddCodeItem(codeEQUOFFSET, n, NotFound, vTyp, tNONE,
    val, 0, o, 0, 0);
end;

procedure AddCodeListEQUSIZE(vTyp: byte; n: string; o: byte; val: TAsmValue);
begin
   AddCodeItem(codeEQUSIZE, n, NotFound, vTyp, tNONE,
    val, 0, o, 0, 0);
end;

procedure AddCodeListSEGOFFSET(vTyp: byte; segIndex: TDefValue;
 o: byte; val: TAsmValue);
begin
   AddCodeItem(codeGETSEGOFFSET, '', segIndex, vTyp, tNONE,
    val, 0, o, 0, 0);
end;

procedure AddCodeListSEGSIZE(vTyp: byte; segIndex: TDefValue;
 o: byte; val: TAsmValue);
begin
   AddCodeItem(codeGETSEGSIZE, '', segIndex, vTyp, tNONE,
    val, 0, o, 0, 0);
end;

procedure AddCodeListRELID(vTyp: byte; id: TDefValue;
 o: byte; val: TAsmValue);
begin
   AddCodeItem(codeRELID, '', id, vTyp, tNONE,
    val, 0, o, 0, 0);
end;

procedure AddCodeListMEMID(vTyp: byte; id: TDefValue;
 o: byte; val: TAsmValue);
begin
   AddCodeItem(codeMEMID, '', id, vTyp, tNONE,
    val, 0, o, 0, 0);
end;



end.
