{ Equates
 (C) 2005-2012, Andreas Berg. All Rights Reserved. }
unit _equ;

interface

uses
 _common;

procedure AddStandardEqus;

procedure SetEquInt(n: string; val: TAsmValue);
procedure SetEquStr(n: string; sOfs: TDefValue; sLen: TAsmValue; fixOfs: boolean);

procedure DefineEqu(n: string);
procedure DefineEquStr(n: string);

implementation

uses
 SysUtils, _error, _mem, _scanner, _parser, _ident, _source;

procedure AddStandardEqus;
begin
   { AddEquInt('?', 0); }
end;



procedure SetEquInt(n: string; val: TAsmValue);
var
 id                : TDefValue;
begin
   id:= GetIdentIndex(n);
   if id = NotFound then begin
      AddEquInt(n, 0);
      id:= GetIdentIndex(n);
   end;
   if IdentList[id].IdentType = itEQU then begin
      IdentList[id].IValue:= val;
      IdentList[id].EquMode:= equModeINT;
      IdentList[id].EquIdentIndex:= Neg;
      IdentList[id].EquOpr:= oprNONE;
      IdentList[id].EquName:= '';
      IdentList[id].SValue:= 0;
      IdentList[id].StrLen:= 0;
      IdentList[id].Len:= 0;
      IdentList[id].RValue:= 0.0;
   end else begin
      ErrorMsg('Identifier is not an Equate [' + n + ']!');
   end;
end;

procedure SetEquStr(n: string; sOfs: TDefValue; sLen: TAsmValue; fixOfs: boolean);
var
 id                : TDefValue;
 uOfs              : TDefValue;
begin
   id:= GetIdentIndex(n);
   MemReplacePos:= -1;
   uOfs:= sOfs;
   if id = NotFound then begin
      AddEquInt(n, 0);
      id:= GetIdentIndex(n);
   end;
   if IdentList[id].IdentType = itEQU then begin
      if (IdentList[id].StrLen > 0) and (not fixOfs) then begin
         if sLen <= IdentList[id].StrLen then begin
            MemReplacePos:= IdentList[id].SValue;
            uOfs:= IdentList[id].SValue;
         end;
      end;
      IdentList[id].IValue:= 0;
      IdentList[id].EquMode:= equModeSTR;
      IdentList[id].EquIdentIndex:= Neg;
      IdentList[id].EquOpr:= oprNONE;
      IdentList[id].EquName:= '';
      IdentList[id].SValue:= uOfs;
      IdentList[id].StrLen:= sLen;
      IdentList[id].Len:= sLen;
      IdentList[id].RValue:= 0.0;
   end else begin
      ErrorMsg('Identifier is not an Equate [' + n + ']!');
   end;
end;



procedure DefineEqu(n: string);
var
 tPos              : integer;
 nam, eName        : string;
 id                : TDefValue;
 val               : TAsmValue;
begin
   SkipBlank;
   tPos:= Source[CurLevel].Pos;
   nam:= GetName;
   if not IsErr then begin
      if nam <> '' then begin
         if (nam = 'OFFSET') or (nam = 'SIZE') then begin
            eName:= GetName;
            if not IsErr then begin
               if eName <> '' then begin
                  id:= GetIdentIndex(eName);
                  if id = NotFound then begin
                     val:= GetCalc;
                     if not IsErr then begin
                        if nam = 'OFFSET' then begin
                           AddEquOffset(n, eName, RetOpr, val);
                        end else begin
                           AddEquSize(n, eName, RetOpr, val);
                        end;
                     end;
                  end else begin
                     Source[CurLevel].Pos:= tPos;
                     val:= GetNumber(tAUTO);
                     if not IsErr then AddEquInt(n, val);
                  end;
               end else begin
                  ErrorMsg('Name expected after [' + nam + ']!');
               end;
            end;
         end else begin
            Source[CurLevel].Pos:= tPos;
            val:= GetNumber(tAUTO);
            if not IsErr then AddEquInt(n, val);
         end;
      end else begin
         val:= GetNumber(tAUTO);
         if not IsErr then AddEquInt(n, val);
      end;
   end;
end;

procedure DefineEquStr(n: string);
var
 strOfs            : TDefValue;
 strLen            : TAsmValue;
 nam, iname, iStr  : string;
 id, i             : TDefValue;
 v, val, sLen, asc : TAsmValue;
 b, t              : byte;
 srcStrOfs         : TDefValue;
 srcStrLen         : TAsmValue;
 eFnd              : boolean;
 vStr, gStr        : string;
begin
   MemReplacePos:= -1;
   strOfs:= MemLength;
   strLen:= 0;

   while true do begin
      nam:= GetName;
      if not IsErr then begin
         if nam <> '' then begin
	          if (nam = 'OFFSET') or (nam = 'SIZE') then begin
               iname:= GetName;
               if not IsErr then begin
                  if iname <> '' then begin
                     id:= GetIdentIndex(iname);
                     if id <> NotFound then begin
                        iStr:= '0';
                        if nam = 'OFFSET' then begin
                           v:= IdentList[id].Offset;
                           val:= GetCalc;
                           v:= DoCalc(v, RetOpr, val);
                           iStr:= IntToStr(v);
                        end else begin
                           v:= IdentList[id].Len;
                           val:= GetCalc;
                           v:= DoCalc(v, RetOpr, val);
                           iStr:= IntToStr(v);
                        end;
                        sLen:= Length(iStr);
                        if sLen > 0 then begin
                           i:= 1;
                           while i <= sLen do begin
                              b:= ord(iStr[i]);
                              AddMemValue(b);
                              Inc(i);
                           end;
                        end;
                        strLen:= strLen + sLen;
                     end else begin
                        ErrorMsg('Identifier not found [' + iname + ']!');
                     end;
                  end else begin
                     ErrorMsg('Name expected after [' + nam + ']!');
                  end;
               end;
            end else begin
               id:= GetIdentIndex(nam);
               if id <> NotFound then begin
                  if IdentList[id].IdentType = itEQU then begin
                     if IdentList[id].EquMode = equModeSTR then begin
                        srcStrOfs:= IdentList[id].SValue;
                        srcStrLen:= IdentList[id].StrLen;
                        if srcStrLen > 0 then begin
                           i:= 0;
                           while i < srcStrLen do begin
                              b:= MemData[srcStrOfs + i];
                              AddMemValue(b);
                              Inc(i);
                           end;
                        end;
                        strLen:= strLen + srcStrLen;
                     end else begin
                        val:= 0;
                        t:= IdentList[id].EquMode;
                        eFnd:= false;
                        if t = equModeINT then begin
                           eFnd:= true;
                           val:= IdentList[id].IValue;
                        end;
                        { Case equModeOFFSET: }
                        { Case equModeSIZE: }

                        if not eFnd then begin
                           ErrorMsg('Incompatible equate type [' + nam + ']!');
                        end;
                        if not IsErr then begin
                           vStr:= IntToStr(val);
                           sLen:= Length(vStr);
                           if sLen > 0 then begin
                              i:= 1;
                              while i <= sLen do begin
                                 b:= ord(vStr[i]);
                                 AddMemValue(b);
                                 Inc(i);
                              end;
                           end;
                           strLen:= strLen + sLen;
                        end;
                     end;
                  end else begin
                     ErrorMsg('Identifier is not an equate [' + nam + ']!');
                  end;
               end else begin
                  ErrorMsg('Identifier not found [' + nam + ']!');
               end;
            end;
         end else begin
            if (GetSrcChar = '"') or (GetSrcChar = #39) then begin
               gStr:= GrabSimpleString;
               if not IsErr then begin
                  sLen:= Length(gStr);
                  if sLen > 0 then begin
                     i:= 1;
                     while i <= sLen do begin
                        b:= ord(gStr[i]);
                        AddMemValue(b);
                        Inc(i);
                     end;
                  end;
                  strLen:= strLen + sLen;
               end;
            end else begin
               if GetSrcChar = '#' then begin
		              NextChar;
                  asc:= GetNum;
                  if (asc < 0) or (asc > 255) then begin
                     ErrorMsg('ASCII out of range (0-255)!');
                  end;
                  if not IsErr then begin
                     AddMemValue(asc);
                     Inc(strLen);
                  end;
               end else begin
                  val:= GetNum;
                  if not IsErr then begin
                     vStr:= IntToStr(val);
                     sLen:= Length(vStr);
                     i:= 1;
                     while i <= sLen do begin
                        b:= ord(vStr[i]);
                        AddMemValue(b);
                        Inc(i);
                     end;
                     strLen:= strLen + sLen;
                  end;
               end;
            end;
         end;
      end;
      if IsErr then break;
      SkipBlank;
      if (GetSrcChar = ',') or (GetSrcChar = '+') then begin
         NextChar;
         SkipBlank;
      end else begin
         break;
      end;
   end;
   if not IsErr then begin
      AddEquString(n, strOfs, strLen);
   end;
end;

end.
