{ Identifiers
 (C) 2005-2012, Andreas Berg. All Rights Reserved. }
unit _ident;

interface

uses
 _common;

const
 lstLOCAL          = 0;
 lstPARENT         = 1;
 lstGLOBAL         = 2;

 imLOCAL           = 0;
 imGLOBAL          = 1;

 itNONE            = 0;
 itLABEL           = 1;
 itMEM             = 2;
 itEQU             = 3;
 itMACRO           = 4;
 itOBJECT          = 5;
 itSECTION         = 6;
 itOTHER           = 255;

 equModeNONE       = 0;
 equModeINT        = 1;
 equModeSTR        = 2;
 equModeOFFSET     = 8;
 equModeSIZE       = 9;
 equModeFLOAT      = 16;

type
 TIdent = record
    Name           : string;
    SourceID       : TDefValue;
    IsDeleted      : boolean;

    IdentType      : byte;
    MemType        : byte;
    Offset         : TAsmValue;
    Len            : TAsmValue;
    EquOpr         : byte;
    EquMode        : byte;
    EquName        : string;
    EquIdentIndex  : TDefValue;
    IValue         : TAsmValue;
    RValue         : extended;
    SValue         : TDefValue;
    StrLen         : TDefValue;
    MacroOffset    : TDefValue;
    MacroLen       : TDefValue;
    ParamOffset    : TDefValue;
    ParamCnt       : TDefValue;
 end;

var
 IdentList         : array of TIdent;
 IdentCount        : TDefValue;
 NeedEquUpdate     : boolean;

 IdentMode         : byte;

procedure InitIdents;
procedure UpdateEquates;

function GetIdentList(n: string): byte;
function ExtractIdentName(n: string): string;

function GetLocal(n: string): TDefValue;
function GetParent(n: string): TDefValue;
function GetGlobal(n: string): TDefValue;
function GetIdentIndex(n: string): TDefValue;

function DeleteIdent(n :string): boolean;

procedure AddIdent(n: string);

procedure AddLabel(n: string; o: TAsmValue);
procedure AddMemory(n: string; o: TAsmValue; typ: byte; siz: TAsmValue);
procedure AddEquInt(n: string; val: TAsmValue);
procedure AddEquString(n: string; sOfs, sLen: TDefValue);
procedure AddEquOffset(n, iname: string; vopr: byte; val: TAsmValue);
procedure AddEquSize(n, iname: string; vopr: byte; val: TAsmValue);
procedure AddMacro(n: string; mOfs, mLen, pOfs, pLen: TDefValue);
procedure AddObject(n: string; mOfs, mLen, pOfs, pLen: TDefValue);
procedure AddSection(n: string; o, s: TAsmValue);
procedure AddDefinedName(n: string);

implementation

uses
 _error, _scanner, _parser, _source, _codegen;

procedure InitIdents;
begin
   SetLength(IdentList, 1);
   IdentCount:= 0;
   IdentMode:= imLOCAL;
end;

procedure CheckEquate1(i: TDefValue);
var
 id                : TDefValue;
 eIndex            : TDefValue;
 eOpr              : byte;
 eInt              : TAsmValue;
 ofs               : TAsmValue;
begin
   if IdentList[i].EquIdentIndex = NotFound then begin
      id:= GetIdentIndex(IdentList[i].EquName);
      if not id = NotFound then begin
         IdentList[i].EquIdentIndex:= id;
      end else begin
         NeedEquUpdate:= true;
      end;
   end;
   if not IdentList[i].EquIdentIndex = NotFound then begin
      eIndex:= IdentList[i].EquIdentIndex;
      eOpr:= IdentList[i].EquOpr;
      eInt:= IdentList[i].IValue;
      ofs:= IdentList[eIndex].Offset;
      ofs:= DoCalc(ofs, eOpr, eInt);
      IdentList[i].EquMode:= equModeINT;
      IdentList[i].IValue:= ofs;
      IdentList[i].EquName:= '';
      IdentList[i].EquIdentIndex:= NotFound;
      IdentList[i].EquOpr:= oprNONE;
   end;
end;

procedure CheckEquate2(i: TDefValue);
var
 id                : TDefValue;
 eIndex            : TDefValue;
 eOpr              : byte;
 eInt              : TAsmValue;
 siz               : TAsmValue;
begin
   if IdentList[i].EquIdentIndex = NotFound then begin
      id:= GetIdentIndex(IdentList[i].EquName);
      if not id = NotFound then begin
         IdentList[i].EquIdentIndex:= id;
      end else begin
         NeedEquUpdate:= true;
      end;
   end;
   if not IdentList[i].EquIdentIndex = NotFound then begin
      eIndex:= IdentList[i].EquIdentIndex;
      eOpr:= IdentList[i].EquOpr;
      eInt:= IdentList[i].IValue;
      siz:= IdentList[eIndex].Len;
      siz:= DoCalc(siz, eOpr, eInt);
      IdentList[i].EquMode:= equModeINT;
      IdentList[i].IValue:= siz;
      IdentList[i].EquName:= '';
      IdentList[i].EquIdentIndex:= NotFound;
      IdentList[i].EquOpr:= oprNONE;
   end;
end;

procedure UpdateEquates;
var
 i                 : TDefValue;
begin
   if NeedEquUpdate then begin
      NeedEquUpdate:= false;
      i:= 0;
      while i < IdentCount do begin
         if IdentList[i].IdentType = itEQU then begin
            case IdentList[i].EquMode of
	           equModeOFFSET: CheckEquate1(i);
               equModeSIZE: CheckEquate2(i);
            end;
         end;
         Inc(i);
      end;
   end;
end;



function GetIdentList(n: string): byte;
var
 retList           : byte;
begin
   retList:= lstLOCAL;
   if Length(n) > 0 then begin
      if n[1] = '~' then begin
         if Length(n) > 1 then begin
            if n[2] = '~' then begin
               retList:= lstGLOBAL;
            end else begin
               retList:= lstPARENT;
            end;
         end;
      end;
   end;
   {if (retList = lstLOCAL) and (IdentMode = imGLOBAL) then retList:= lstGLOBAL;}
   GetIdentList:= retList;
end;

function ExtractIdentName(n: string): string;
var
 nn                : string;
 i                 : integer;
begin
   nn:= '';
   if Length(n) > 0 then begin
      for i:= 1 to Length(n) do begin
         if n[i] <> '~' then nn:= nn + n[i];
      end;
   end;
   ExtractIdentName:= nn;
end;



function GetLocal(n: string): TDefValue;
var
 ind, sID, i       : TDefValue;
begin
   ind:= NotFound;
   sID:= Source[CurLevel].SourceID;
   if IdentCount > 0 then begin
      i:= 0;
      while i < IdentCount do begin
         if (IdentList[i].Name = n) and (not IdentList[i].IsDeleted) and
          (IdentList[i].SourceID = sID) then begin
            ind:= i;
            break;
         end;
         Inc(i);
      end;
   end;
   GetLocal:= ind;
end;

function GetParent(n: string): TDefValue;
var
 ind, fID, i       : TDefValue;
begin
   ind:= NotFound;
   fID:= Source[FileLevel].SourceID;
   if IdentCount > 0 then begin
      i:= 0;
      while i < IdentCount do begin
         if (IdentList[i].Name = n) and (not IdentList[i].IsDeleted) and
          (IdentList[i].SourceID = fID) then begin
            ind:= i;
            break;
         end;
         Inc(i);
      end;
   end;
   GetParent:= ind;
end;

function GetGlobal(n: string): TDefValue;
var
 ind, i            : TDefValue;
begin
   ind:= NotFound;
   if IdentCount > 0 then begin
      i:= 0;
      while i < IdentCount do begin
         if (IdentList[i].Name = n) and (not IdentList[i].IsDeleted) and
          (IdentList[i].SourceID = 0) then begin
            ind:= i;
            break;
         end;
         Inc(i);
      end;
   end;
   GetGlobal:= ind;
end;

function GetIdentIndex(n: string): TDefValue;
var
 retIndex          : TDefValue;
 useList           : byte;
 nn                : string;
begin
   useList:= GetIdentList(n);
   nn:= ExtractIdentName(n);
   if useList = lstLOCAL then begin
      {if IdentMode = imLOCAL then begin }
         retIndex:= GetLocal(nn);
         if retIndex = NotFound then retIndex:= GetParent(nn);
         if retIndex = NotFound then retIndex:= GetGlobal(nn);
      {end else begin
         retIndex:= GetGlobal(nn);
         if retIndex = NotFound then retIndex:= GetLocal(nn);
         if retIndex = NotFound then retIndex:= GetParent(nn);
      end;}
   end else begin
      if useList = lstPARENT then begin
         retIndex:= GetParent(nn);
      end else begin
         retIndex:= GetGlobal(nn);
      end;
   end;
   GetIdentIndex:= retIndex;
end;



function DeleteIdent(n :string): boolean;
var
 delStatus         : boolean;
 useList           : byte;
 nn, iname         : string;
 i, srcID          : TDefValue;
begin
   delStatus:= false;
   useList:= GetIdentList(n);
   nn:= ExtractIdentName(n);
   srcID:= 0;
   if useList = lstLOCAL then srcID:= Source[CurLevel].SourceID;
   if useList = lstPARENT then srcID:= Source[FileLevel].SourceID;
   {if (useList = lstLOCAL) and (IdentMode = imGLOBAL) then srcID:= 0;}
   if IdentCount > 0 then begin
      i:= 0;
      while i < IdentCount do begin
         iname:= IdentList[i].Name;
         if nn = Copy(iname, 1, Length(nn)) then begin
            if (nn = iname) or ((iname[Length(nn)] = '.') and
             (Length(iname) > Length(nn))) then begin
               if IdentList[i].SourceID = srcID then begin
                  if not IdentList[i].IsDeleted then begin
                     UpdateCodeID;
                     IdentList[i].IsDeleted:= true;
                     delStatus:= true;
                  end;
               end;
            end;
         end;
         Inc(i);
      end;
   end;
   DeleteIdent:= delStatus;
end;



procedure AddIdent(n: string);
var
 useList           : byte;
 nn                : string;
 srcID             : TDefValue;
begin
   if GetIdentIndex(n) = NotFound then begin
      useList:= GetIdentList(n);
      nn:= ExtractIdentName(n);
      srcID:= 0;
      If useList = lstLOCAL then srcID:= Source[CurLevel].SourceID;
      If useList = lstPARENT then srcID:= Source[FileLevel].SourceID;
      if (useList = lstLOCAL) and (IdentMode = imGLOBAL) then srcID:= 0;
      SetLength(IdentList, IdentCount + 1);
      IdentList[IdentCount].Name:= nn;
      IdentList[IdentCount].SourceID:= srcID;
      IdentList[IdentCount].IsDeleted:= false;

      IdentList[IdentCount].IdentType:= itNONE;
      IdentList[IdentCount].MemType:= tNONE;
      IdentList[IdentCount].Offset:= 0;
      IdentList[IdentCount].Len:= 0;
      IdentList[IdentCount].EquOpr:= oprNONE;
      IdentList[IdentCount].EquMode:= equModeNONE;
      IdentList[IdentCount].EquName:= '';
      IdentList[IdentCount].EquIdentIndex:= NotFound;
      IdentList[IdentCount].IValue:= 0;
      IdentList[IdentCount].RValue:= 0.0;
      IdentList[IdentCount].SValue:= 0;
      IdentList[IdentCount].StrLen:= 0;
      IdentList[IdentCount].MacroOffset:= 0;
      IdentList[IdentCount].MacroLen:= 0;
      IdentList[IdentCount].ParamOffset:= 0;
      IdentList[IdentCount].ParamCnt:= 0;
      Inc(IdentCount);
   end else begin
      ErrorMsg('Identifier already exists [' + n + ']!');
   end;
end;



procedure AddLabel(n: string; o: TAsmValue);
begin
   AddIdent(n);
   if not IsErr then begin
      IdentList[IdentCount-1].IdentType:= itLABEL;
      IdentList[IdentCount-1].Offset:= o;
   end;
end;

procedure AddMemory(n: string; o: TAsmValue; typ: byte; siz: TAsmValue);
begin
   AddIdent(n);
   if not IsErr then begin
      IdentList[IdentCount-1].IdentType:= itMEM;
      IdentList[IdentCount-1].Offset:= o;
      IdentList[IdentCount-1].MemType:= typ;
      IdentList[IdentCount-1].Len:= siz;
   end;
end;

procedure AddEquInt(n: string; val: TAsmValue);
begin
   AddIdent(n);
   if not IsErr then begin
      IdentList[IdentCount-1].IdentType:= itEQU;
      IdentList[IdentCount-1].IValue:= val;
      IdentList[IdentCount-1].EquMode:= equModeINT;
   end;
end;

procedure AddEquString(n: string; sOfs, sLen: TDefValue);
begin
   AddIdent(n);
   if not IsErr then begin
      IdentList[IdentCount-1].IdentType:= itEQU;
      IdentList[IdentCount-1].SValue:= sOfs;
      IdentList[IdentCount-1].StrLen:= sLen;
      IdentList[IdentCount-1].EquMode:= equModeSTR;
      IdentList[IdentCount-1].Len:= sLen;
   end;
end;

procedure AddEquOffset(n, iname: string; vopr: byte; val: TAsmValue);
begin
   AddIdent(n);
   if not IsErr then begin
      IdentList[IdentCount-1].IdentType:= itEQU;
      IdentList[IdentCount-1].EquName:= iname;
      IdentList[IdentCount-1].EquOpr:= vopr;
      IdentList[IdentCount-1].IValue:= val;
      IdentList[IdentCount-1].EquMode:= equModeOFFSET;
   end;
end;

procedure AddEquSize(n, iname: string; vopr: byte; val: TAsmValue);
begin
   AddIdent(n);
   if not IsErr then begin
      IdentList[IdentCount-1].IdentType:= itEQU;
      IdentList[IdentCount-1].EquName:= iname;
      IdentList[IdentCount-1].EquOpr:= vopr;
      IdentList[IdentCount-1].IValue:= val;
      IdentList[IdentCount-1].EquMode:= equModeSIZE;
   end;
end;

procedure AddMacro(n: string; mOfs, mLen, pOfs, pLen: TDefValue);
var
 nn                : string;
begin
   nn:= ExtractIdentName(n);
   AddIdent('~~' + nn);
   if not IsErr then begin
      IdentList[IdentCount-1].IdentType:= itMACRO;
      IdentList[IdentCount-1].MacroOffset:= mOfs;
      IdentList[IdentCount-1].MacroLen:= mLen;
      IdentList[IdentCount-1].ParamOffset:= pOfs;
      IdentList[IdentCount-1].ParamCnt:= pLen;
   end;
end;

procedure AddObject(n: string; mOfs, mLen, pOfs, pLen: TDefValue);
var
 nn                : string;
begin
   nn:= ExtractIdentName(n);
   AddIdent('~~' + nn);
   if not IsErr then begin
      IdentList[IdentCount-1].IdentType:= itOBJECT;
      IdentList[IdentCount-1].MacroOffset:= mOfs;
      IdentList[IdentCount-1].MacroLen:= mLen;
      IdentList[IdentCount-1].ParamOffset:= pOfs;
      IdentList[IdentCount-1].ParamCnt:= pLen;
   end;
end;

procedure AddSection(n: string; o, s: TAsmValue);
begin
   AddIdent(n);
   if not IsErr then begin
      IdentList[IdentCount-1].IdentType:= itSECTION;
      IdentList[IdentCount-1].Offset:= o;
      IdentList[IdentCount-1].Len:= s;
   end;
end;

procedure AddDefinedName(n: string);
begin
   AddIdent(n);
   if not IsErr then begin
      IdentList[IdentCount-1].IdentType:= itOTHER;
   end;
end;

end.
 
