{ PIE (Processor Instruction Engine)
 (C) 2005-2012, Andreas Berg. All Rights Reserved. }
unit _pie;

interface

uses
 _common;

const
 iDataLen          = 32;

 pNone             = 0;
 pReg8             = 1;
 pReg16            = 2;
 pReg32            = 3;
 pSegReg           = 4;
 pCtrlReg          = 5;
 pDebugReg         = 6;
 pTestReg          = 7;
 pRegAL            = 8;  { special }
 pRegAX            = 9;  { special }
 pRegEAX           = 10; { special }
 pRegCL            = 11; { special }
 pRegDX            = 12; { special }
 pRegSI            = 13; { special }
 pRegDI            = 14; { special }
 pRegBX            = 15; { special }
 pBYTE_SI          = 16;
 pBYTE_DI          = 17;
 pBYTE_BX          = 18;
 pWORD_SI          = 19;
 pWORD_DI          = 20;
 pWORD_BX          = 21;
 pDWORD_SI         = 22;
 pDWORD_DI         = 23;
 pDWORD_BX         = 24;
 pRegES            = 25; { special }
 pRegCS            = 26; { special }
 pRegSS            = 27; { special }
 pRegDS            = 28; { special }
 pRegFS            = 29; { special }
 pRegGS            = 30; { special }
 pES_SI            = 31;
 pES_DI            = 32;
 pES_BX            = 33;
 pCS_SI            = 34;
 pCS_DI            = 35;
 pCS_BX            = 36;
 pSS_SI            = 37;
 pSS_DI            = 38;
 pSS_BX            = 39;
 pDS_SI            = 40;
 pDS_DI            = 41;
 pDS_BX            = 42;
 pFS_SI            = 43;
 pFS_DI            = 44;
 pFS_BX            = 45;
 pGS_SI            = 46;
 pGS_DI            = 47;
 pGS_BX            = 48;

 pES_memory        = 49;
 pCS_memory        = 50;
 pSS_memory        = 51;
 pDS_memory        = 52;
 pFS_memory        = 53;
 pGS_memory        = 54;

 pES_value         = 55;
 pCS_value         = 56;
 pSS_value         = 57;
 pDS_value         = 58;
 pFS_value         = 59;
 pGS_value         = 60;

 pName             = 61;
 pBYTEName         = 62;
 pWORDName         = 63;
 pDWORDName        = 64;

 pValue            = 65;
 pBYTEValue        = 66;
 pWORDValue        = 67;
 pDWORDValue       = 68;
 pValue1           = 69; { special }

 pCODE             = 70;
 pDATA             = 71;
 pSSEG             = 72;
 pBSS              = 73;
 pFSEG             = 74;
 pGSEG             = 75;

 pINS              = 76;
 pINSB             = 77;
 pINSW             = 78;
 pINSD             = 79;
 pMOVS             = 80;
 pMOVSB            = 81;
 pMOVSW            = 82;
 pMOVSD            = 83;
 pOUTS             = 84;
 pOUTSB            = 85;
 pOUTSW            = 86;
 pOUTSD            = 87;
 pLODS             = 88;
 pLODSB            = 89;
 pLODSW            = 90;
 pLODSD            = 91;
 pSTOS             = 92;
 pSTOSB            = 93;
 pSTOSW            = 94;
 pSTOSD            = 95;
 pCMPS             = 96;
 pCMPSB            = 97;
 pCMPSW            = 98;
 pCMPSD            = 99;
 pSCAS             = 100;
 pSCASB            = 101;
 pSCASW            = 102;
 pSCASD            = 103;

 pBX_SI            = 104;
 pBX_DI            = 105;
 pBP_SI            = 106;
 pBP_DI            = 107;
 pBX_SI_MEM        = 108;
 pBX_DI_MEM        = 109;
 pBP_SI_MEM        = 110;
 pBP_DI_MEM        = 111;
 pSI_MEM           = 112;
 pDI_MEM           = 113;
 pBP_MEM           = 114;
 pBX_MEM           = 115;

 pBX_SI_VALUE      = 116;
 pBX_DI_VALUE      = 117;
 pBP_SI_VALUE      = 118;
 pBP_DI_VALUE      = 119;
 pSI_VALUE         = 120;
 pDI_VALUE         = 121;
 pBP_VALUE         = 122;
 pBX_VALUE         = 123;

 pRegDL            = 124; { special }
 pRegBL            = 125; { special }
 pRegAH            = 126; { special }
 pRegCH            = 127; { special }
 pRegDH            = 128; { special }
 pRegBH            = 129; { special }

 pRegCX            = 130; { special }
 pRegSP            = 131; { special }
 pRegBP            = 132; { special }

 pRegECX           = 133; { special }
 pRegEDX           = 134; { special }
 pRegEBX           = 135; { special }
 pRegESP           = 136; { special }
 pRegEBP           = 137; { special }
 pRegESI           = 138; { special }
 pRegEDI           = 139; { special }

 pRegCR0           = 140; { special }
 pRegCR2           = 141; { special }
 pRegCR3           = 142; { special }
 pRegCR4           = 143; { special }

 pRegDR0           = 144; { special }
 pRegDR1           = 145; { special }
 pRegDR2           = 146; { special }
 pRegDR3           = 147; { special }
 pRegDR6           = 148; { special }
 pRegDR7           = 149; { special }

 pRegTR3           = 150; { special }
 pRegTR4           = 151; { special }
 pRegTR5           = 152; { special }
 pRegTR6           = 153; { special }
 pRegTR7           = 154; { special }

 pPtrSI            = 155;
 pPtrDI            = 156;
 pPtrBX            = 157;

 pPtrName          = 158;
 pPtrValue         = 159;

 pES_BX_SI         = 160;
 pES_BX_DI         = 161;
 pES_BP_SI         = 162;
 pES_BP_DI         = 163;
 pES_BX_SI_MEM     = 164;
 pES_BX_DI_MEM     = 165;
 pES_BP_SI_MEM     = 166;
 pES_BP_DI_MEM     = 167;
 pES_SI_MEM        = 168;
 pES_DI_MEM        = 169;
 pES_BP_MEM        = 170;
 pES_BX_MEM        = 171;
 pES_BX_SI_VALUE   = 172;
 pES_BX_DI_VALUE   = 173;
 pES_BP_SI_VALUE   = 174;
 pES_BP_DI_VALUE   = 175;
 pES_SI_VALUE      = 176;
 pES_DI_VALUE      = 177;
 pES_BP_VALUE      = 178;
 pES_BX_VALUE      = 179;

 pCS_BX_SI         = 180;
 pCS_BX_DI         = 181;
 pCS_BP_SI         = 182;
 pCS_BP_DI         = 183;
 pCS_BX_SI_MEM     = 184;
 pCS_BX_DI_MEM     = 185;
 pCS_BP_SI_MEM     = 186;
 pCS_BP_DI_MEM     = 187;
 pCS_SI_MEM        = 188;
 pCS_DI_MEM        = 189;
 pCS_BP_MEM        = 190;
 pCS_BX_MEM        = 191;
 pCS_BX_SI_VALUE   = 192;
 pCS_BX_DI_VALUE   = 193;
 pCS_BP_SI_VALUE   = 194;
 pCS_BP_DI_VALUE   = 195;
 pCS_SI_VALUE      = 196;
 pCS_DI_VALUE      = 197;
 pCS_BP_VALUE      = 198;
 pCS_BX_VALUE      = 199;

 pSS_BX_SI         = 200;
 pSS_BX_DI         = 201;
 pSS_BP_SI         = 202;
 pSS_BP_DI         = 203;
 pSS_BX_SI_MEM     = 204;
 pSS_BX_DI_MEM     = 205;
 pSS_BP_SI_MEM     = 206;
 pSS_BP_DI_MEM     = 207;
 pSS_SI_MEM        = 208;
 pSS_DI_MEM        = 209;
 pSS_BP_MEM        = 210;
 pSS_BX_MEM        = 211;
 pSS_BX_SI_VALUE   = 212;
 pSS_BX_DI_VALUE   = 213;
 pSS_BP_SI_VALUE   = 214;
 pSS_BP_DI_VALUE   = 215;
 pSS_SI_VALUE      = 216;
 pSS_DI_VALUE      = 217;
 pSS_BP_VALUE      = 218;
 pSS_BX_VALUE      = 219;

 pDS_BX_SI         = 220;
 pDS_BX_DI         = 221;
 pDS_BP_SI         = 222;
 pDS_BP_DI         = 223;
 pDS_BX_SI_MEM     = 224;
 pDS_BX_DI_MEM     = 225;
 pDS_BP_SI_MEM     = 226;
 pDS_BP_DI_MEM     = 227;
 pDS_SI_MEM        = 228;
 pDS_DI_MEM        = 229;
 pDS_BP_MEM        = 230;
 pDS_BX_MEM        = 231;
 pDS_BX_SI_VALUE   = 232;
 pDS_BX_DI_VALUE   = 233;
 pDS_BP_SI_VALUE   = 234;
 pDS_BP_DI_VALUE   = 235;
 pDS_SI_VALUE      = 236;
 pDS_DI_VALUE      = 237;
 pDS_BP_VALUE      = 238;
 pDS_BX_VALUE      = 239;

 pFS_BX_SI         = 240;
 pFS_BX_DI         = 241;
 pFS_BP_SI         = 242;
 pFS_BP_DI         = 243;
 pFS_BX_SI_MEM     = 244;
 pFS_BX_DI_MEM     = 245;
 pFS_BP_SI_MEM     = 246;
 pFS_BP_DI_MEM     = 247;
 pFS_SI_MEM        = 248;
 pFS_DI_MEM        = 249;
 pFS_BP_MEM        = 250;
 pFS_BX_MEM        = 251;
 pFS_BX_SI_VALUE   = 252;
 pFS_BX_DI_VALUE   = 253;
 pFS_BP_SI_VALUE   = 254;
 pFS_BP_DI_VALUE   = 255;
 pFS_SI_VALUE      = 256;
 pFS_DI_VALUE      = 257;
 pFS_BP_VALUE      = 258;
 pFS_BX_VALUE      = 259;

 pGS_BX_SI         = 260;
 pGS_BX_DI         = 261;
 pGS_BP_SI         = 262;
 pGS_BP_DI         = 263;
 pGS_BX_SI_MEM     = 264;
 pGS_BX_DI_MEM     = 265;
 pGS_BP_SI_MEM     = 266;
 pGS_BP_DI_MEM     = 267;
 pGS_SI_MEM        = 268;
 pGS_DI_MEM        = 269;
 pGS_BP_MEM        = 270;
 pGS_BX_MEM        = 271;
 pGS_BX_SI_VALUE   = 272;
 pGS_BX_DI_VALUE   = 273;
 pGS_BP_SI_VALUE   = 274;
 pGS_BP_DI_VALUE   = 275;
 pGS_SI_VALUE      = 276;
 pGS_DI_VALUE      = 277;
 pGS_BP_VALUE      = 278;
 pGS_BX_VALUE      = 279;

 pValue7           = 280; { special }

 pSegOffset        = 300; { 0000:0000 }

type
 TInstr = record
    Name           : string;
    par1,par2      : word;
    iData          : string[iDataLen+1];
 end;



var
 InstrList         : array of TInstr;
 InstrCount        : cardinal;

 paValue           : array[1..2] of TAsmValue;
 paString          : array[1..2] of string;
 paType            : array[1..2] of word;
 paSpecial         : array[1..2] of word;
 paOpr             : array[1..2] of byte;

 CodeUsed          : boolean;
 RunInstr2         : boolean;


procedure InitInstr;
procedure GrabInstr;

procedure FixInstr32;
procedure RemoveInstr32;
procedure DumpInstr(fil: string);

procedure RunInstr(iinstr: string);

implementation

uses
 SysUtils, _conv, _error, _ident, _scanner, _func,
 _parser, _source, _codegen, _reg2, _seg, _cpu, _header;


procedure InitInstr;
begin
   SetLength(InstrList, 1);
   InstrCount:= 0;
end;

procedure GrabInstr;
var
 n                 : string;
 idt               : string;
 p1                : TAsmValue;
 p2                : TAsmValue;
begin
   n:= GetName;
   if n <> '' then begin
      p1:= GetNum;
      if not IsErr then begin
         p2:= GetNum;
         if not IsErr then begin
            idt:= GetData;
            if length(idt) = idataLen then begin
               SetLength(InstrList, InstrCount+1);
               InstrList[InstrCount].Name:= n;
               InstrList[InstrCount].Par1:= p1;
               InstrList[InstrCount].Par2:= p2;
               InstrList[InstrCount].idata:= idt;
               Inc(InstrCount);
            end else begin
               if idt = '' then begin
                  ErrorMsg('Instruction data expected!');
               end else begin
                  ErrorMsg('Invalid length of instruction data!');
               end;
            end;
         end;
      end;
   end else begin
      ErrorMsg('Name expected!');
   end;
end;



procedure FixInstr32;
begin
   { Disabled! }
end;

procedure RemoveInstr32;
var
 i                 : TDefValue;
 istr              : string;
begin
   if InstrCount > 0 then begin
      i:= 0;
      while i < InstrCount do begin
         istr:= InstrList[i].iData;
         istr[7]:= '0';
         istr[9]:= '0';
         istr[6]:= '0';
         istr[8]:= '0';
         InstrList[i].iData:= istr;
         Inc(i);
      end;
   end;
end;

procedure DumpInstr(fil: string);
var
 f                 : textfile;
 i                 : TDefValue;
 n, d              : string;
 p1, p2            : word;
 s                 : string;
begin
   {$I-}
   Assign(f, fil);
   Rewrite(f);
      if InstrCount > 0 then begin
         i:= 0;
         while i < InstrCount do begin
            n:= InstrList[i].Name;
            p1:= InstrList[i].Par1;
            p2:= InstrList[i].Par2;
            d:= InstrList[i].iData;
            s:= 'instr ' + InsertBlanks(n, 12) + '  ';
            s:= s + InsertBlanks(IntToStr(p1), 3) + '  ';
            s:= s + InsertBlanks(IntToStr(p2), 3) + '  ';
            s:= s + d;
            WriteLn(f, s);
            Inc(i);
         end;
         WriteLn(f, '');
         WriteLn(f, 'end');
      end;
   Close(f);
   {$I+}
end;



procedure GrabParameter(pIndex: byte);
label
 getVal, endCheck, endProc;
var
 n, n2, n3         : string;
 v,v2              : TAsmValue;
 s                 : word;
 t                 : word;
 tmp               : cardinal;
 r                 : string;
 id                : TDefValue;
 ts                : string;
 o                 : byte;
 segr              : string;
begin
   n:= ''; v:= 0; s:= 0; t:= 0; o:= 0;
   SkipBlank;
   if IsENDLine then goto endCheck;
   tmp:= Source[CurLevel].Pos;

   if not IsENDLine then begin
      if GetSrcChar = '[' then begin
         NextChar;
         SkipBlank;
         n:= GetName;
         if n <> '' then begin
            SkipBlank;
            if GetSrcChar = ':' then begin
               NextChar;
               SkipBlank;
               if IsSegReg(n) then begin
                  segr:= n;
                  n:= GetName;
                  if n <> '' then begin
                     if not IsAnyReg(n) then begin
                        if segr = 'ES' then t:= pES_memory;
                        if segr = 'CS' then t:= pCS_memory;
                        if segr = 'SS' then t:= pSS_memory;
                        if segr = 'DS' then t:= pDS_memory;
                        if segr = 'FS' then t:= pFS_memory;
                        if segr = 'GS' then t:= pGS_memory;
                        id:= GetIdentIndex(n);
                        if (IdentList[id].IdentType = itEQU) and
                         (IdentList[id].EquMode = equModeINT) then begin
                           if segr = 'ES' then t:= pES_value;
                           if segr = 'CS' then t:= pCS_value;
                           if segr = 'SS' then t:= pSS_value;
                           if segr = 'DS' then t:= pDS_value;
                           if segr = 'FS' then t:= pGS_value;
                           if segr = 'GS' then t:= pGS_value;
                        end;
                        v:= GetCalc;
                        if IsErr then goto endCheck;
                        o:= retOpr;
                     end else begin
                        if (n = 'SI') or (n = 'DI') or (n = 'BX') or (n = 'BP') then begin
                           if n = 'SI' then begin
                              if segr = 'ES' then t:= pES_SI;
                              if segr = 'CS' then t:= pCS_SI;
                              if segr = 'SS' then t:= pSS_SI;
                              if segr = 'DS' then t:= pDS_SI;
                              if segr = 'FS' then t:= pFS_SI;
                              if segr = 'GS' then t:= pGS_SI;
                              SkipBlank;
                              if GetSrcChar = '+' then begin
                                 NextChar;
                                 n2:= GetName;
                                 if n2 <> '' then begin
                                    if not IsAnyReg(n2) then begin
                                       if segr = 'ES' then t:= pES_SI_MEM;
                                       if segr = 'CS' then t:= pCS_SI_MEM;
                                       if segr = 'SS' then t:= pSS_SI_MEM;
                                       if segr = 'DS' then t:= pDS_SI_MEM;
                                       if segr = 'FS' then t:= pFS_SI_MEM;
                                       if segr = 'GS' then t:= pGS_SI_MEM;
                                       id:= GetIdentIndex(n2);
                                       if (IdentList[id].IdentType = itEQU) and
                                        (IdentList[id].EquMode = equModeINT) then begin
                                          if segr = 'ES' then t:= pES_SI_VALUE;
                                          if segr = 'CS' then t:= pCS_SI_VALUE;
                                          if segr = 'SS' then t:= pSS_SI_VALUE;
                                          if segr = 'DS' then t:= pDS_SI_VALUE;
                                          if segr = 'FS' then t:= pFS_SI_VALUE;
                                          if segr = 'GS' then t:= pGS_SI_VALUE;
                                       end;
                                       v:= GetCalc;
                                       if IsErr then goto endCheck;
                                       o:= retOpr;
                                       n:= n2;
                                    end;
                                 end else begin
                                    ErrorMsg('Memory or Value expected after "+"!');
                                    goto endCheck;
                                 end;
                              end;
                           end;
                        end;
                        if n = 'DI' then begin
                           if segr = 'ES' then t:= pES_DI;
                           if segr = 'CS' then t:= pCS_DI;
                           if segr = 'SS' then t:= pSS_DI;
                           if segr = 'DS' then t:= pDS_DI;
                           if segr = 'FS' then t:= pFS_DI;
                           if segr = 'GS' then t:= pGS_DI;
                           SkipBlank;
                           if GetSrcChar = '+' then begin
                              NextChar;
                              n2:= GetName;
                              if n2 <> '' then begin
                                 if not IsAnyReg(n2) then begin
                                    if segr = 'ES' then t:= pES_DI_MEM;
                                    if segr = 'CS' then t:= pCS_DI_MEM;
                                    if segr = 'SS' then t:= pSS_DI_MEM;
                                    if segr = 'DS' then t:= pDS_DI_MEM;
                                    if segr = 'FS' then t:= pFS_DI_MEM;
                                    if segr = 'GS' then t:= pGS_DI_MEM;
                                    id:= GetIdentIndex(n2);
                                    if (IdentList[id].IdentType = itEQU) and
                                     (IdentList[id].EquMode = equModeINT) then begin
                                       if segr = 'ES' then t:= pES_DI_VALUE;
                                       if segr = 'CS' then t:= pCS_DI_VALUE;
                                       if segr = 'SS' then t:= pSS_DI_VALUE;
                                       if segr = 'DS' then t:= pDS_DI_VALUE;
                                       if segr = 'FS' then t:= pFS_DI_VALUE;
                                       if segr = 'GS' then t:= pGS_DI_VALUE;
                                    end;
                                    v:= GetCalc;
                                    if IsErr then goto endCheck;
                                    o:= retOpr;
                                    n:= n2;
                                 end;
                              end else begin
                                 ErrorMsg('Memory or Value expected after "+"!');
                                 goto endCheck;
                              end;
                           end;
                        end;
                        if n = 'BX' then begin
                           if segr = 'ES' then t:= pES_BX;
                           if segr = 'CS' then t:= pCS_BX;
                           if segr = 'SS' then t:= pSS_BX;
                           if segr = 'DS' then t:= pDS_BX;
                           if segr = 'FS' then t:= pFS_BX;
                           if segr = 'GS' then t:= pGS_BX;
                           SkipBlank;
                           if GetSrcChar = '+' then begin
                              NextChar;
                              n2:= GetName;
                              if n2 <> '' then begin
                                 if (n2 = 'SI') or (n2 = 'DI') then begin
                                    if n2 = 'SI' then begin
                                       if segr = 'ES' then t:= pES_BX_SI;
                                       if segr = 'CS' then t:= pCS_BX_SI;
                                       if segr = 'SS' then t:= pSS_BX_SI;
                                       if segr = 'DS' then t:= pDS_BX_SI;
                                       if segr = 'FS' then t:= pFS_BX_SI;
                                       if segr = 'GS' then t:= pGS_BX_SI;
                                    end else begin
                                       if segr = 'ES' then t:= pES_BX_DI;
                                       if segr = 'CS' then t:= pCS_BX_DI;
                                       if segr = 'SS' then t:= pSS_BX_DI;
                                       if segr = 'DS' then t:= pDS_BX_DI;
                                       if segr = 'FS' then t:= pFS_BX_DI;
                                       if segr = 'GS' then t:= pGS_BX_DI;
                                    end;
                                    SkipBlank;
                                    if GetSrcChar = '+' then begin
                                       NextChar;
                                       n3:= GetName;
                                       if n3 <> '' then begin
                                          if n2 = 'SI' then begin
                                             if segr = 'ES' then t:= pES_BX_SI_MEM;
                                             if segr = 'CS' then t:= pCS_BX_SI_MEM;
                                             if segr = 'SS' then t:= pSS_BX_SI_MEM;
                                             if segr = 'DS' then t:= pDS_BX_SI_MEM;
                                             if segr = 'FS' then t:= pFS_BX_SI_MEM;
                                             if segr = 'GS' then t:= pGS_BX_SI_MEM;
                                          end else begin
                                             if segr = 'ES' then t:= pES_BX_DI_MEM;
                                             if segr = 'CS' then t:= pCS_BX_DI_MEM;
                                             if segr = 'SS' then t:= pSS_BX_DI_MEM;
                                             if segr = 'DS' then t:= pDS_BX_DI_MEM;
                                             if segr = 'FS' then t:= pFS_BX_DI_MEM;
                                             if segr = 'GS' then t:= pGS_BX_DI_MEM;
                                          end;
                                          id:= GetIdentIndex(n3);
                                          if (IdentList[id].IdentType = itEQU) and
                                           (IdentList[id].EquMode = equModeINT) then begin
                                             if n2 = 'SI' then begin
                                                if segr = 'ES' then t:= pES_BX_SI_VALUE;
                                                if segr = 'CS' then t:= pCS_BX_SI_VALUE;
                                                if segr = 'SS' then t:= pSS_BX_SI_VALUE;
                                                if segr = 'DS' then t:= pDS_BX_SI_VALUE;
                                                if segr = 'FS' then t:= pFS_BX_SI_VALUE;
                                                if segr = 'GS' then t:= pGS_BX_SI_VALUE;
                                             end else begin
                                                if segr = 'ES' then t:= pES_BX_DI_VALUE;
                                                if segr = 'CS' then t:= pCS_BX_DI_VALUE;
                                                if segr = 'SS' then t:= pSS_BX_DI_VALUE;
                                                if segr = 'DS' then t:= pDS_BX_DI_VALUE;
                                                if segr = 'FS' then t:= pFS_BX_DI_VALUE;
                                                if segr = 'GS' then t:= pGS_BX_DI_VALUE;
                                             end;
                                          end;
                                          v:= GetCalc;
                                          if IsErr then goto endCheck;
                                          o:= retOpr;
                                          n:= n3;
                                       end else begin
                                          ErrorMsg('Memory or Value expected after [ + ]!');
                                          goto endCheck;
                                       end;
                                    end;
                                 end else begin
                                    if not IsAnyReg(n2) then begin
                                       if segr = 'ES' then t:= pES_BX_MEM;
                                       if segr = 'CS' then t:= pCS_BX_MEM;
                                       if segr = 'SS' then t:= pSS_BX_MEM;
                                       if segr = 'DS' then t:= pDS_BX_MEM;
                                       if segr = 'FS' then t:= pFS_BX_MEM;
                                       if segr = 'GS' then t:= pGS_BX_MEM;
                                       id:= GetIdentIndex(n2);
                                       if (IdentList[id].IdentType = itEQU) and
                                        (IdentList[id].EquMode = equModeINT) then begin
                                          if segr = 'ES' then t:= pES_BX_VALUE;
                                          if segr = 'CS' then t:= pCS_BX_VALUE;
                                          if segr = 'SS' then t:= pSS_BX_VALUE;
                                          if segr = 'DS' then t:= pDS_BX_VALUE;
                                          if segr = 'FS' then t:= pFS_BX_VALUE;
                                          if segr = 'GS' then t:= pGS_BX_VALUE;
                                       end;
                                       v:= GetCalc;
                                       if IsErr then goto endCheck;
                                       o:= retOpr;
                                       n:= n2;
                                    end else begin
                                       ErrorMsg('SI, DI, Memory or Value expected after [ + ]!');
                                       goto endCheck;
                                    end;
                                 end;
                              end else begin
                                 ErrorMsg('Pointer-Register, Memory or Value expected after [ + ]!');
                                 goto endCheck;
                              end;
                           end;
                        end;
                        if n = 'BP' then begin
                           SkipBlank;
                           if GetSrcChar = '+' then begin
                              NextChar;
                              n2:= GetName;
                              if n2 <> '' then begin
                                 if (n2 = 'SI') or (n2 = 'DI') then begin
                                    if n2 = 'SI' then begin
                                       if segr = 'ES' then t:= pES_BP_SI;
                                       if segr = 'CS' then t:= pCS_BP_SI;
                                       if segr = 'SS' then t:= pSS_BP_SI;
                                       if segr = 'DS' then t:= pDS_BP_SI;
                                       if segr = 'FS' then t:= pFS_BP_SI;
                                       if segr = 'GS' then t:= pGS_BP_SI;
                                    end else begin
                                       if segr = 'ES' then t:= pES_BP_DI;
                                       if segr = 'CS' then t:= pCS_BP_DI;
                                       if segr = 'SS' then t:= pSS_BP_DI;
                                       if segr = 'DS' then t:= pDS_BP_DI;
                                       if segr = 'FS' then t:= pFS_BP_DI;
                                       if segr = 'GS' then t:= pGS_BP_DI;
                                    end;
                                    SkipBlank;
                                    if GetSrcChar = '+' then begin
                                       NextChar;
                                       n3:= GetName;
                                       if n3 <> '' then begin
                                          if n2 = 'SI' then begin
                                             if segr = 'ES' then t:= pES_BP_SI_MEM;
                                             if segr = 'CS' then t:= pCS_BP_SI_MEM;
                                             if segr = 'SS' then t:= pSS_BP_SI_MEM;
                                             if segr = 'DS' then t:= pDS_BP_SI_MEM;
                                             if segr = 'FS' then t:= pFS_BP_SI_MEM;
                                             if segr = 'GS' then t:= pGS_BP_SI_MEM;
                                          end else begin
                                             if segr = 'ES' then t:= pES_BP_DI_MEM;
                                             if segr = 'CS' then t:= pCS_BP_DI_MEM;
                                             if segr = 'SS' then t:= pSS_BP_DI_MEM;
                                             if segr = 'DS' then t:= pDS_BP_DI_MEM;
                                             if segr = 'FS' then t:= pFS_BP_DI_MEM;
                                             if segr = 'GS' then t:= pGS_BP_DI_MEM;
                                          end;
                                          id:= GetIdentIndex(n3);
                                          if (IdentList[id].IdentType = itEQU) and
                                           (IdentList[id].EquMode = equModeINT) then begin
                                             if n2 = 'SI' then begin
                                                if segr = 'ES' then t:= pES_BP_SI_VALUE;
                                                if segr = 'CS' then t:= pCS_BP_SI_VALUE;
                                                if segr = 'SS' then t:= pSS_BP_SI_VALUE;
                                                if segr = 'DS' then t:= pDS_BP_SI_VALUE;
                                                if segr = 'FS' then t:= pFS_BP_SI_VALUE;
                                                if segr = 'GS' then t:= pGS_BP_SI_VALUE;
                                             end else begin
                                                if segr = 'ES' then t:= pES_BP_DI_VALUE;
                                                if segr = 'CS' then t:= pCS_BP_DI_VALUE;
                                                if segr = 'SS' then t:= pSS_BP_DI_VALUE;
                                                if segr = 'DS' then t:= pDS_BP_DI_VALUE;
                                                if segr = 'FS' then t:= pFS_BP_DI_VALUE;
                                                if segr = 'GS' then t:= pGS_BP_DI_VALUE;
                                             end;
                                          end;
                                          v:= GetCalc;
                                          if IsErr then goto endCheck;
                                          o:= retOpr;
                                          n:= n3;
                                       end else begin
                                          ErrorMsg('Memory or Value expected after [ + ]!');
                                          goto endCheck;
                                       end;
                                    end;
                                 end else begin
                                    if not IsAnyReg(n2) then begin
                                       if segr = 'ES' then t:= pES_BP_MEM;
                                       if segr = 'CS' then t:= pCS_BP_MEM;
                                       if segr = 'SS' then t:= pSS_BP_MEM;
                                       if segr = 'DS' then t:= pDS_BP_MEM;
                                       if segr = 'FS' then t:= pFS_BP_MEM;
                                       if segr = 'GS' then t:= pGS_BP_MEM;
                                       id:= GetIdentIndex(n2);
                                       if (IdentList[id].IdentType = itEQU) and
                                        (IdentList[id].EquMode = equModeINT) then begin
                                          if segr = 'ES' then t:= pES_BP_VALUE;
                                          if segr = 'CS' then t:= pCS_BP_VALUE;
                                          if segr = 'SS' then t:= pSS_BP_VALUE;
                                          if segr = 'DS' then t:= pDS_BP_VALUE;
                                          if segr = 'FS' then t:= pFS_BP_VALUE;
                                          if segr = 'GS' then t:= pGS_BP_VALUE;
                                       end;
                                       v:= GetCalc;
                                       if IsErr then goto endCheck;
                                       o:= retOpr;
                                       n:= n2;
                                    end else begin
                                       ErrorMsg('SI, DI, Memory or Value expected after [ + ]!');
                                       goto endCheck;
                                    end;
                                 end;
                              end else begin
                                 ErrorMsg('Pointer-Register, Memory or Value expected after [ + ]!');
                                 goto endCheck;
                              end;
                           end;
                        end;
                     end;
                  end else begin
                     v:= GetNumber(tDWORD);
                     if IsErr then goto endCheck;
                     o:= retOpr;
                     if segr = 'ES' then t:= pES_VALUE;
                     if segr = 'CS' then t:= pCS_VALUE;
                     if segr = 'SS' then t:= pSS_VALUE;
                     if segr = 'DS' then t:= pDS_VALUE;
                     if segr = 'FS' then t:= pFS_VALUE;
                     if segr = 'GS' then t:= pGS_VALUE;
                  end;
               end else begin
                  ErrorMsg('Segment register expected before [ : ]!');
                  goto endCheck;
               end;
            end else begin
               if (n = 'SI') or (n = 'DI') or (n = 'BX') or (n = 'BP') then begin
                  if n = 'SI' then begin
                     t:= pPtrSI;
                     SkipBlank;
                     if GetSrcChar = '+' then begin
                        NextChar;
                        n2:= GetName;
                        if n2 <> '' then begin
                           if not IsAnyReg(n2) then begin
                              t:= pSI_MEM;
                              id:= GetIdentIndex(n2);
                              if (IdentList[id].IdentType = itEQU) and
                              (IdentList[id].EquMode = equModeINT) then t:= pSI_VALUE;
                              v:= GetCalc;
                              if IsErr then goto endCheck;
                              o:= retOpr;
                              n:= n2;
                           end;
                        end else begin
                           ErrorMsg('Memory or Value expected after [ + ]!');
                           goto endCheck;
                        end;
                     end;
                  end;
                  if n = 'DI' then begin
                     t:= pPtrDI;
                     SkipBlank;
                     if GetSrcChar = '+' then begin
                        NextChar;
                        n2:= GetName;
                        if n2 <> '' then begin
                           if not IsAnyReg(n2) then begin
                              t:= pDI_MEM;
                              id:= GetIdentIndex(n2);
                              if (IdentList[id].IdentType = itEQU) and
                               (IdentList[id].EquMode = equModeINT) then t:= pDI_VALUE;
                              v:= GetCalc;
                              if IsErr then goto endCheck;
                              o:= retOpr;
                              n:= n2;
                           end;
                        end else begin
                           ErrorMsg('Memory or Value expected after [ + ]!');
                           goto endCheck;
                        end;
                     end;
                  end;
                  if n = 'BX' then begin
                     t:= pPtrBX;
                     SkipBlank;
                     if GetSrcChar = '+' then begin
                        NextChar;
                        n2:= GetName;
                        if n2 <> '' then begin
                           if (n2 = 'SI') or (n2 = 'DI') then begin
                              if n2 = 'SI' then t:= pBX_SI;
                              if n2 = 'DI' then t:= pBX_DI;
                              SkipBlank;
                              if GetSrcChar = '+' then begin
                                 NextChar;
                                 n3:= GetName;
                                 if n3 <> '' then begin
                                    if n2 = 'SI' then t:= pBX_SI_MEM;
                                    if n2 = 'DI' then t:= pBX_DI_MEM;
                                    id:= GetIdentIndex(n2);
                                    if ((IdentList[id].IdentType = itEQU) and
                                     (IdentList[id].EquMode = equModeINT)) and
                                     (n2 = 'SI') then t:= pBX_SI_VALUE;
                                    if ((IdentList[id].IdentType = itEQU) and
                                     (IdentList[id].EquMode = equModeINT)) and
                                     (n2 = 'DI') then t:= pBX_DI_VALUE;
                                    v:= GetCalc;
                                    if IsErr then goto endCheck;
                                    o:= retOpr;
                                    n:= n3;
                                 end else begin
                                    ErrorMsg('Memory or Value expected after [ + ]!');
                                    goto endCheck;
                                 end;
                              end;
                           end else begin
                              if not IsAnyReg(n2) then begin
                                 t:= pBX_MEM;
                                 id:= GetIdentIndex(n2);
                                 if (IdentList[id].IdentType = itEQU) and
                                 (IdentList[id].EquMode = equModeINT) then t:= pBX_VALUE;
                                 v:= GetCalc;
                                 if IsErr then goto endCheck;
                                 o:= retOpr;
                                 n:= n2;
                              end else begin
                                 ErrorMsg('SI, DI, Memory or Value expected after [ + ]!');
                                 goto endCheck;
                              end;
                           end;
                        end else begin
                           ErrorMsg('Pointer-Register, Memory or Value expected after [ + ]!');
                           goto endCheck;
                        end;
                     end;
                  end;
                  if n = 'BP' then begin
                     SkipBlank;
                     if GetSrcChar = '+' then begin
                        NextChar;
                        n2:= GetName;
                        if n2 <> '' then begin
                           if (n2 = 'SI') or (n2 = 'DI') then begin
                              if n2 = 'SI' then t:= pBP_SI;
                              if n2 = 'DI' then t:= pBP_DI;
                              SkipBlank;
                              if GetSrcChar = '+' then begin
                                 NextChar;
                                 n3:= GetName;
                                 if n3 <> '' then begin
                                    if n2 = 'SI' then t:= pBP_SI_MEM;
                                    if n2 = 'DI' then t:= pBP_DI_MEM;
                                    id:= GetIdentIndex(n2);
                                    if ((IdentList[id].IdentType = itEQU) and
                                     (IdentList[id].EquMode = equModeINT)) and
                                     (n2 = 'SI') then t:= pBP_SI_VALUE;
                                    if ((IdentList[id].IdentType = itEQU) and
                                     (IdentList[id].EquMode = equModeINT)) and
                                     (n2 = 'DI') then t:= pBP_DI_VALUE;
                                    v:= GetCalc;
                                    if IsErr then goto endCheck;
                                    o:= retOpr;
                                    n:= n3;
                                 end else begin
                                    ErrorMsg('Memory or Value expected after [ + ]!');
                                    goto endCheck;
                                 end;
                              end;
                           end else begin
                              if not IsAnyReg(n2) then begin
                                 t:= pBP_MEM;
                                 id:= GetIdentIndex(n2);
                                 if (IdentList[id].IdentType = itEQU) and
                                 (IdentList[id].EquMode = equModeINT) then t:= pBP_VALUE;
                                 v:= GetCalc;
                                 if IsErr then goto endCheck;
                                 o:= retOpr;
                                 n:= n2;
                              end else begin
                                 ErrorMsg('SI, DI, Memory or Value expected after [ + ]!');
                                 goto endCheck;
                              end;
                           end;
                        end else begin
                           ErrorMsg('Pointer-Register, Memory or Value expected after [ + ]!');
                           goto endCheck;
                        end;
                     end;
                  end;
               end else begin
                  if not IsAnyReg(n) then begin
                     t:= pPtrName;
                     id:= GetIdentIndex(n);
                     if id <> NotFound then begin
                        if (IdentList[id].IdentType = itEQU) and
                         (IdentList[id].EquMode = equModeINT) then t:= pPtrValue;
                     end;
                     v:= GetCalc;
                     if IsErr then goto endCheck;
                     o:= retOpr;
                  end else begin
                     ErrorMsg('Segment-, pointer-register or memory expected!');
                     goto endCheck;
                  end;
               end;
            end;
         end else begin
            ErrorMsg('Register/Name expected between [ [] ]!');
            goto endCheck;
         end;
         SkipBlank;
         if GetSrcChar = ']' then begin
            NextChar;
         end else begin
            ErrorMsg('[ ] ] expected!');
         end;
         goto endCheck;
      end;

      n:= GetName;
      if n <> '' then begin
         if n = 'OFFSET' then begin
            n:= GetName;
            if n <> '' then begin
               id:= GetIdentIndex(n);
               if id <> NotFound then begin
                  v:= IdentList[id].Offset;
                  v2:= GetCalc;
                  if retOpr > 0 then v:= DoCalc(v, retOpr, v2);
                  t:= pValue;
                  if v = 1 then s:= pValue1;
                  if v = 7 then s:= pValue7;
                  goto endCheck;
               end else begin
                  ErrorMsg('Identifier not found [' + n + ']!');
                  goto endCheck;
               end;
            end else begin
               ErrorMsg('Name expected!');
            end;
            goto endCheck;
         end;

         if n = 'VALUE' then begin
            n:= GetName;
            if n <> '' then begin
               id:= GetIdentIndex(n);
               if id <> NotFound then begin
                  v:= IdentList[id].IValue;
                  v2:= GetCalc;
                  if retOpr > 0 then v:= DoCalc(v, retOpr, v2);
                  t:= pValue;
                  if v = 1 then s:= pValue1;
                  if v = 7 then s:= pValue7;
                  goto endCheck;
               end else begin
                  ErrorMsg('Identifier not found [' + n + ']!');
                  goto endCheck;
               end;
            end else begin
               ErrorMsg('Name expected!');
               goto endCheck;
            end;
         end;

         if n = 'SIZE' then begin
            n:= GetName;
            if n <> '' then begin
               id:= GetIdentIndex(n);
               if id <> NotFound then begin
                  v:= IdentList[id].Len;
                  v2:= GetCalc;
                  if retOpr > 0 then v:= DoCalc(v, retOpr, v2);
                  t:= pValue;
                  if v = 1 then s:= pValue1;
                  if v = 7 then s:= pValue7;
                  goto endCheck;
               end else begin
                  ErrorMsg('Identifier not found [' + n + ']!');
                  goto endCheck;
               end;
            end else begin
               ErrorMsg('Name expected!');
               goto endCheck;
            end;
         end;

         id:= GetIdentIndex(n);
         if id <> NotFound then begin
            if (IdentList[id].IdentType = itEQU) then begin
               if (IdentList[id].EquMode = equModeINT)
                or (IdentList[id].EquMode = equModeSTR) then begin
                  n:= ''; goto getVal;
               end;
            end;
         end;

         if n = '@CODE' then t:= pCode;
         if n = '@DATA' then t:= pData;
         if n = '@BSS' then t:= pBss;
         if n = '@SSEG' then t:= pSSEG;
         if n = '@FSEG' then t:= pFSEG;
         if n = '@GSEG' then t:= pGSEG;

         if n = 'INS' then t:= pINS;
         if n = 'INSB' then t:= pINSB;
         if n = 'INSW' then t:= pINSW;
         if n = 'INSD' then t:= pINSD;
         if n = 'MOVS' then t:= pMOVS;
         if n = 'MOVSB' then t:= pMOVSB;
         if n = 'MOVSW' then t:= pMOVSW;
         if n = 'MOVSD' then t:= pMOVSD;
         if n = 'OUTS' then t:= pOUTS;
         if n = 'OUTSB' then t:= pOUTSB;
         if n = 'OUTSW' then t:= pOUTSW;
         if n = 'OUTSD' then t:= pOUTSD;
         if n = 'LODS' then t:= pLODS;
         if n = 'LODSB' then t:= pLODSB;
         if n = 'LODSW' then t:= pLODSW;
         if n = 'LODSD' then t:= pLODSD;
         if n = 'STOS' then t:= pSTOS;
         if n = 'STOSB' then t:= pSTOSB;
         if n = 'STOSW' then t:= pSTOSW;
         if n = 'STOSD' then t:= pSTOSD;
         if n = 'CMPS' then t:= pCMPS;
         if n = 'CMPSB' then t:= pCMPSB;
         if n = 'CMPSW' then t:= pCMPSW;
         if n = 'CMPSD' then t:= pCMPSD;
         if n = 'SCAS' then t:= pSCAS;
         if n = 'SCASB' then t:= pSCASB;
         if n = 'SCASW' then t:= pSCASW;
         if n = 'SCASD' then t:= pSCASD;
         if t <> 0 then goto endCheck;

         if n = 'BYTE' then begin
            t:= 0;
            IgnoreChar('[');
            n:= GetName;
            if n <> '' then begin
               if n = 'PTR' then begin
                  IgnoreChar('[');
                  n:= GetName;
                  if (n <> 'SI') and (n <> 'DI') and (n <> 'BX') then begin
                     ErrorMsg('SI, DI or BX expected!');
                     goto endCheck;
                  end;
               end;
               IgnoreChar(']');
               if n = 'SI' then begin
                  t:= pBYTE_SI;
               end else begin
                  if n = 'DI' then begin
                     t:= pBYTE_DI;
                  end else begin
                     if n = 'BX' then begin
                        t:= pBYTE_BX;
                     end else begin
                        t:= pBYTEName;
                        v:= GetCalc;
                        o:= retOpr;
                        IgnoreChar(']');
                        if IsErr then goto endCheck;
                     end;
                  end;
               end;
            end else begin
               v:= GetNumber(tBYTE);
               if IsErr then goto endCheck;
               t:= pBYTEValue;
               if v = 1 then s:= pValue1;
               if v = 7 then s:= pValue7;
               IgnoreChar(']');
            end;
            goto endCheck;
         end;
         if n = 'WORD' then begin
            t:= 0;
            IgnoreChar('[');
            n:= GetName;
            if n <> '' then begin
               if n = 'PTR' then begin
                  IgnoreChar('[');
                  n:= GetName;
                  if (n <> 'SI') and (n <> 'DI') and (n <> 'BX') then begin
                     ErrorMsg('SI, DI or BX expected!');
                     goto endCheck;
                  end;
               end;
               IgnoreChar(']');
               if n = 'SI' then begin
                  t:= pWORD_SI;
               end else begin
                  if n = 'DI' then begin
                     t:= pWORD_DI;
                  end else begin
                     if n = 'BX' then begin
                        t:= pWORD_BX;
                     end else begin
                        t:= pWORDName;
                        v:= GetCalc;
                        o:= retOpr;
                        IgnoreChar(']');
                        if IsErr then goto endCheck;
                     end;
                  end;
               end;
            end else begin
               v:= GetNumber(tWORD);
               if IsErr then goto endCheck;
               t:= pWORDValue;
               if v = 1 then s:= pValue1;
               if v = 7 then s:= pValue7;
               IgnoreChar(']');
            end;
            goto endCheck;
         end;
         if n = 'DWORD' then begin
            t:= 0;
            IgnoreChar('[');
            n:= GetName;
            if n <> '' then begin
               if n = 'PTR' then begin
                  IgnoreChar('[');
                  n:= GetName;
                  if (n <> 'SI') and (n <> 'DI') and (n <> 'BX') and
                   (n <> 'ESI') and (n <> 'EDI') and (n <> 'EBX') then begin
                     ErrorMsg('SI, DI or BX expected!');
                     goto endCheck;
                  end;
               end;
               IgnoreChar(']');
               if (n = 'SI') or (n = 'ESI') then begin
                  t:= pDWORD_SI;
               end else begin
                  if (n = 'DI') or (n = 'EDI') then begin
                     t:= pDWORD_DI;
                  end else begin
                     if (n = 'BX') or (n = 'EBX') then begin
                        t:= pDWORD_BX;
                     end else begin
                        t:= pDWORDName;
                        v:= GetCalc;
                        o:= retOpr;
                        IgnoreChar(']');
                        if IsErr then goto endCheck;
                     end;
                  end;
               end;
            end else begin
               v:= GetNumber(tDWORD);
               if IsErr then goto endCheck;
               t:= pDWORDValue;
               if v = 1 then s:= pValue1;
               if v = 7 then s:= pValue7;
               IgnoreChar(']');
            end;
            goto endCheck;
         end;

         if IsSegReg(n) then begin
            t:= pSegReg;
            if n = 'ES' then s:= pRegES;
            if n = 'CS' then s:= pRegCS;
            if n = 'SS' then s:= pRegSS;
            if n = 'DS' then s:= pRegDS;
            if n = 'FS' then s:= pRegFS;
            if n = 'GS' then s:= pRegGS;
            SkipBlank;
            if not IsENDLine then begin
               if GetSrcChar = ':' then begin
                  NextChar;
                  SkipBlank;
                  r:= n;
                  IgnoreChar('[');
                  n:= GetName;
                  if n <> '' then begin
                     IgnoreChar(']');
                     if (n = 'SI') or (n = 'ESI') then begin
                        if r = 'ES' then t:= pES_SI;
                        if r = 'CS' then t:= pCS_SI;
                        if r = 'SS' then t:= pSS_SI;
                        if r = 'DS' then t:= pDS_SI;
                        if r = 'FS' then t:= pFS_SI;
                        if r = 'GS' then t:= pGS_SI;
                        s:= 0;
                     end else begin
                        if (n = 'DI') or (n = 'EDI') then begin
                           if r = 'ES' then t:= pES_DI;
                           if r = 'CS' then t:= pCS_DI;
                           if r = 'SS' then t:= pSS_DI;
                           if r = 'DS' then t:= pDS_DI;
                           if r = 'FS' then t:= pFS_DI;
                           if r = 'GS' then t:= pGS_DI;
                           s:= 0;
                        end else begin
                           if (n = 'BX') or (n = 'EBX') then begin
                              if r = 'ES' then t:= pES_BX;
                              if r = 'CS' then t:= pCS_BX;
                              if r = 'SS' then t:= pSS_BX;
                              if r = 'DS' then t:= pDS_BX;
                              if r = 'FS' then t:= pFS_BX;
                              if r = 'GS' then t:= pGS_BX;
                              s:= 0;
                           end else begin
                              id:= GetIdentIndex(n);
                              if id <> NotFound then begin
                                 if (IdentList[id].IdentType = itEQU)
                                  and (IdentList[id].EquMode = equModeINT) then begin
                                    v:= IdentList[id].IValue;
                                    if r = 'ES' then t:= pES_value;
                                    if r = 'CS' then t:= pCS_value;
                                    if r = 'SS' then t:= pSS_value;
                                    if r = 'DS' then t:= pDS_value;
                                    if r = 'FS' then t:= pFS_value;
                                    if r = 'GS' then t:= pGS_value;
                                    s:= 0;
                                    v2:= GetCalc;
                                    o:= retOpr;
                                    v:= DoCalc(v, o, v2);
                                    IgnoreChar(']');
                                    goto endCheck;
                                 end;
                              end;
                              if r = 'ES' then t:= pES_memory;
                              if r = 'CS' then t:= pCS_memory;
                              if r = 'SS' then t:= pSS_memory;
                              if r = 'DS' then t:= pDS_memory;
                              if r = 'FS' then t:= pFS_memory;
                              if r = 'GS' then t:= pGS_memory;
                              s:= 0;
                              v:= GetCalc;
                              o:= retOpr;
                              IgnoreChar(']');
                              if IsErr then goto endCheck;
                           end;
                        end;
                     end;
                  end else begin
                     v:= GetNumber(tAUTO);
                     if IsErr then goto endCheck;
                     if r = 'ES' then t:= pES_value;
                     if r = 'CS' then t:= pCS_value;
                     if r = 'SS' then t:= pSS_value;
                     if r = 'DS' then t:= pDS_value;
                     if r = 'FS' then t:= pFS_value;
                     if r = 'GS' then t:= pGS_value;
                     s:= 0;
                     IgnoreChar(']');
                  end;
                  goto endCheck;
               end;
            end;
            goto endCheck;
         end;

         if IsReg8(n) then begin
            t:= pReg8;
            if n = 'AL' then s:= pRegAL;
            if n = 'CL' then s:= pRegCL;
            if n = 'DL' then s:= pRegDL;
            if n = 'BL' then s:= pRegBL;
            if n = 'AH' then s:= pRegAH;
            if n = 'CH' then s:= pRegCH;
            if n = 'DH' then s:= pRegDH;
            if n = 'BH' then s:= pRegBH;
            goto endCheck;
         end;
         if IsReg16(n) then begin
            t:= pReg16;
            if n = 'AX' then s:= pRegAX;
            if n = 'DX' then s:= pRegDX;
            if n = 'SI' then begin
               s:= pRegSI;
               SkipBlank;
               if not IsENDLine then begin
                  if GetSrcChar = '+' then begin
                     NextChar;
                     ts:= GetName;
                     if ts <> '' then begin
                        id:= GetIdentIndex(ts);
                        if id <> NotFound then begin
                           if (IdentList[id].IdentType = itEQU) and
                            (IdentList[id].EquMode = equModeINT) then begin
                              v:= IdentList[id].IValue;
                              v2:= GetCalc;
                              o:= retOpr;
                              v:= DoCalc(v, o, v2);
                              if IsErr then goto endCheck;
                              t:= pSI_VALUE;
                              s:= pNone;
                              goto endCheck;
                           end;
                        end;
                        t:= pSI_MEM;
                        v:= GetCalc;
                        o:= retOpr;
                        if IsErr then goto endCheck;
                        s:= pNone;
                        n:= ts;
                        goto endCheck;
                     end else begin
                        v:= GetNumber(tAUTO);
                        if IsErr then goto endCheck;
                        t:= pSI_VALUE;
                        s:= pNone;
                        goto endCheck;
                     end;
                  end;
               end;
            end;
            if n = 'DI' then begin
               s:= pRegDI;
               SkipBlank;
               if not IsENDLine then begin
                  if GetSrcChar = '+' then begin
                     NextChar;
                     ts:= GetName;
                     if ts <> '' then begin
                        id:= GetIdentIndex(ts);
                        if id <> NotFound then begin
                           if (IdentList[id].IdentType = itEQU) and
                            (IdentList[id].EquMode = equModeINT) then begin
                              v:= IdentList[id].IValue;
                              v2:= GetCalc;
                              o:= retOpr;
                              v:= DoCalc(v, o, v2);
                              if IsErr then goto endCheck;
                              t:= pDI_VALUE;
                              s:= pNone;
                              goto endCheck;
                           end;
                        end;
                        t:= pDI_MEM;
                        v:= GetCalc;
                        o:= retOpr;
                        if IsErr then goto endCheck;
                        s:= pNone;
                        n:= ts;
                        goto endCheck;
                     end else begin
                        v:= GetNumber(tAUTO);
                        if IsErr then goto endCheck;
                        t:= pDI_VALUE;
                        s:= pNone;
                        goto endCheck;
                     end;
                  end;
               end;
            end;
            if n = 'BP' then begin
               s:= pRegBP;
               SkipBlank;
               if not IsENDLine then begin
                  if GetSrcChar = '+' then begin
                     NextChar;
                     ts:= GetName;
                     if ts <> '' then begin
                        if ts = 'SI' then begin
                           t:= pBP_SI;
                           s:= pNone;
                           SkipBlank;
                           if not IsENDLine then begin
                              if GetSrcChar = '+' then begin
                                 NextChar;
                                 ts:= GetName;
                                 if ts <> '' then begin
                                    id:= GetIdentIndex(ts);
                                    if id <> NotFound then begin
                                       if (IdentList[id].IdentType = itEQU) and
                                        (IdentList[id].EquMode = equModeINT) then begin
                                          v:= IdentList[id].IValue;
                                          v2:= GetCalc;
                                          o:= retOpr;
                                          v:= DoCalc(v, o, v2);
                                          if IsErr then goto endCheck;
                                          t:= pBP_SI_VALUE;
                                          s:= pNone;
                                          goto endCheck;
                                       end;
                                    end;
                                    t:= pBP_SI_MEM;
                                    v:= GetCalc;
                                    o:= retOpr;
                                    if IsErr then goto endCheck;
                                    s:= pNone;
                                    n:= ts;
                                 end else begin
                                    v:= GetNumber(tAUTO);
                                    if IsErr then goto endCheck;
                                    v:= IdentList[id].IValue;
                                    t:= pBP_SI_VALUE;
                                    s:= pNone;
                                    goto endCheck;
                                 end;
                              end;
                           end;
                           goto endCheck;
                        end;
                        if ts = 'DI' then begin
                           t:= pBP_DI;
                           s:= pNone;
                           SkipBlank;
                           if not IsENDLine then begin
                              if GetSrcChar = '+' then begin
                                 NextChar;
                                 ts:= GetName;
                                 if ts <> '' then begin
                                    id:= GetIdentIndex(ts);
                                    if id <> NotFound then begin
                                       if (IdentList[id].IdentType = itEQU) and
                                        (IdentList[id].EquMode = equModeINT) then begin
                                          v:= IdentList[id].IValue;
                                          v2:= GetCalc;
                                          o:= retOpr;
                                          v:= DoCalc(v, o, v2);
                                          if IsErr then goto endCheck;
                                          t:= pBP_DI_VALUE;
                                          s:= pNone;
                                          goto endCheck;
                                       end;
                                    end;
                                    t:= pBP_DI_MEM;
                                    v:= GetCalc;
                                    o:= retOpr;
                                    if IsErr then goto endCheck;
                                    s:= pNone;
                                    n:= ts;
                                 end else begin
                                    v:= GetNumber(tAUTO);
                                    if IsErr then goto endCheck;
                                    v:= IdentList[id].IValue;
                                    t:= pBP_DI_VALUE;
                                    s:= pNone;
                                    goto endCheck;
                                 end;
                              end;
                           end;
                           goto endCheck;
                        end;
                        id:= GetIdentIndex(ts);
                        if id <> NotFound then begin
                           if (IdentList[id].IdentType = itEQU) and
                            (IdentList[id].EquMode = equModeINT) then begin
                              v:= IdentList[id].IValue;
                              v2:= GetCalc;
                              o:= retOpr;
                              v:= DoCalc(v, o, v2);
                              if IsErr then goto endCheck;
                              t:= pBP_VALUE;
                              s:= pNone;
                              goto endCheck;
                           end;
                        end;
                        t:= pBP_MEM;
                        v:= GetCalc;
                        o:= retOpr;
                        if IsErr then goto endCheck;
                        s:= pNone;
                        n:= ts;
                        goto endCheck;
                     end else begin
                        v:= GetNumber(tAUTO);
                        if IsErr then goto endCheck;
                        t:= pBP_VALUE;
                        s:= pNone;
                        goto endCheck;
                     end;
                  end;
               end;
            end;
            if n = 'BX' then begin
               s:= pRegBX;
               SkipBlank;
               if not IsENDLine then begin
                  if GetSrcChar = '+' then begin
                     NextChar;
                     ts:= GetName;
                     if ts <> '' then begin
                        if ts = 'SI' then begin
                           t:= pBX_SI;
                           s:= pNone;
                           SkipBlank;
                           if not IsENDLine then begin
                              if GetSrcChar = '+' then begin
                                 NextChar;
                                 ts:= GetName;
                                 if ts <> '' then begin
                                    id:= GetIdentIndex(ts);
                                    if id <> NotFound then begin
                                       if (IdentList[id].IdentType = itEQU) and
                                        (IdentList[id].EquMode = equModeINT) then begin
                                          v:= IdentList[id].IValue;
                                          v2:= GetCalc;
                                          o:= retOpr;
                                          v:= DoCalc(v, o, v2);
                                          if IsErr then goto endCheck;
                                          t:= pBX_SI_VALUE;
                                          s:= pNone;
                                          goto endCheck;
                                       end;
                                    end;
                                    t:= pBX_SI_MEM;
                                    v:= GetCalc;
                                    o:= retOpr;
                                    if IsErr then goto endCheck;
                                    s:= pNone;
                                    n:= ts;
                                 end else begin
                                    v:= GetNumber(tAUTO);
                                    if IsErr then goto endCheck;
                                    v:= IdentList[id].IValue;
                                    t:= pBX_SI_VALUE;
                                    s:= pNone;
                                    goto endCheck;
                                 end;
                              end;
                           end;
                           goto endCheck;
                        end;
                        if ts = 'DI' then begin
                           t:= pBX_DI;
                           s:= pNone;
                           SkipBlank;
                           if not IsENDLine then begin
                              if GetSrcChar = '+' then begin
                                 NextChar;
                                 ts:= GetName;
                                 if ts <> '' then begin
                                    id:= GetIdentIndex(ts);
                                    if id <> NotFound then begin
                                       if (IdentList[id].IdentType = itEQU) and
                                        (IdentList[id].EquMode = equModeINT) then begin
                                          v:= IdentList[id].IValue;
                                          v2:= GetCalc;
                                          o:= retOpr;
                                          v:= DoCalc(v, o, v2);
                                          if IsErr then goto endCheck;
                                          t:= pBX_DI_VALUE;
                                          s:= pNone;
                                          goto endCheck;
                                       end;
                                    end;
                                    t:= pBX_DI_MEM;
                                    v:= GetCalc;
                                    o:= retOpr;
                                    if IsErr then goto endCheck;
                                    s:= pNone;
                                    n:= ts;
                                 end else begin
                                    v:= GetNumber(tAUTO);
                                    if IsErr then goto endCheck;
                                    v:= IdentList[id].IValue;
                                    t:= pBX_DI_VALUE;
                                    s:= pNone;
                                    goto endCheck;
                                 end;
                              end;
                           end;
                           goto endCheck;
                        end;
                        id:= GetIdentIndex(ts);
                        if id <> NotFound then begin
                           if (IdentList[id].IdentType = itEQU) and
                            (IdentList[id].EquMode = equModeINT) then begin
                              v:= IdentList[id].IValue;
                              v2:= GetCalc;
                              o:= retOpr;
                              v:= DoCalc(v, o, v2);
                              if IsErr then goto endCheck;
                              t:= pBX_VALUE;
                              s:= pNone;
                              goto endCheck;
                           end;
                        end;
                        t:= pBX_MEM;
                        v:= GetCalc;
                        o:= retOpr;
                        if IsErr then goto endCheck;
                        s:= pNone;
                        n:= ts;
                        goto endCheck;
                     end else begin
                        v:= GetNumber(tAUTO);
                        if IsErr then goto endCheck;
                        t:= pBX_VALUE;
                        s:= pNone;
                        goto endCheck;
                     end;
                  end;
               end;
            end;
            if n = 'CX' then s:= pRegCX;
            if n = 'SP' then s:= pRegSP;
            goto endCheck;
         end;
         if IsReg32(n) then begin
            t:= pReg32;
            if n = 'EAX' then s:= pRegEAX;
            if n = 'ECX' then s:= pRegECX;
            if n = 'EDX' then s:= pRegEDX;
            if n = 'EBX' then s:= pRegEBX;
            if n = 'ESP' then s:= pRegESP;
            if n = 'EBP' then s:= pRegEBP;
            if n = 'ESI' then s:= pRegESI;
            if n = 'EDI' then s:= pRegEDI;
            goto endCheck;
         end;
         if IsCtrlReg(n) then begin
            t:= pCtrlReg;
            if n = 'CR0' then s:= pRegCR0;
            if n = 'CR2' then s:= pRegCR2;
            if n = 'CR3' then s:= pRegCR3;
            if n = 'CR4' then s:= pRegCR4;
            goto endCheck;
         end;
         if IsDebugReg(n) then begin
            t:= pDebugReg;
            if n = 'DR0' then s:= pRegDR0;
            if n = 'DR1' then s:= pRegDR1;
            if n = 'DR2' then s:= pRegDR2;
            if n = 'DR3' then s:= pRegDR3;
            if n = 'DR6' then s:= pRegDR6;
            if n = 'DR7' then s:= pRegDR7;
            goto endCheck;
         end;
         if IsTestReg(n) then begin
            t:= pTestReg;
            if n = 'TR3' then s:= pRegTR3;
            if n = 'TR4' then s:= pRegTR4;
            if n = 'TR5' then s:= pRegTR5;
            if n = 'TR6' then s:= pRegTR6;
            if n = 'TR7' then s:= pRegTR7;
            goto endCheck;
         end;

         t:= pName;
         v:= GetCalc;
         o:= retOpr;
         goto endCheck;

      end else begin
         goto getVal;
      end;
   end;

getVal:
   Source[CurLevel].Pos:= tmp;

   SkipBlank;
   if not IsENDLine then begin
      if (GetSrcChar = '-') or (IsNumPrefix(GetSrcChar)) or
       (IsHex(GetSrcChar)) or (GetSrcChar = #39) or (IsNameChar(GetSrcChar)) or
       (GetSrcChar = '$') or (GetSrcChar = '(') or (GetSrcChar = '"') then begin
         v:= GetNumber(tAUTO);
         if IsErr then goto endCheck;
         t:= pValue;
         if v = 1 then s:= pValue1;
         if v = 7 then s:= pValue7;
         SkipBlank;
         if not IsEndLine then begin
            if GetSrcChar = ':' then begin
               if pIndex = 2 then begin
                  ErrorMsg('[seg:offset] can NOT be used with another parameter!');
                  goto endProc;
               end;
               NextChar;
               paValue[2]:= v;
               s:= 0;
               v:= GetNumber(tAUTO);
               if IsErr then goto endCheck;
               paValue[1]:= v;
               paType[1]:= pSegOffset;
               paType[2]:= pNONE;
               paSpecial[1]:= 0;
               paSpecial[2]:= 0;
               paOpr[1]:= o;
               paOpr[2]:= 0;
               paString[1]:= n;
               paString[2]:= '';
               goto endProc;
            end;
         end;
      end;
   end;

endCheck:
   paString[pIndex]:= n;
   paValue[pIndex]:= v;
   paType[pIndex]:= t;
   paSpecial[pIndex]:= s;
   paOpr[pIndex]:= o;
endProc:
end;

procedure BuildCode(c1, c2: char);
label
 skip;
var
 b1,b2: byte;
begin
   if (c1 = '-') and (c2 = '-') then goto skip;
   if (c1 = '-') then begin
      if IsHex(c2) then begin
         b2:= GetHexValue(c2);
         AddCodeListVALUE(tBYTE, b2);
      end else begin
         ErrorMsg('Invalid instruction data!');
      end;
   end else begin
      if (c2 = '-') then begin
         if IsHex(c1) then begin
            b1:= GetHexValue(c1);
            AddCodeListVALUE(tBYTE, b1);
         end else begin
            ErrorMsg('Invalid instruction data!');
         end;
      end else begin
         if (IsHex(c1)) and (IsHex(c2)) then begin
            b1:= GetHexValue(c1);
            b2:= GetHexValue(c2);
            AddCodeListVALUE(tBYTE, (b1*16)+b2);
         end else begin
            ErrorMsg('Invalid instruction data!');
         end;
      end;
   end;
skip:
end;

function GetCode(c1, c2: char): byte;
label
 skip;
var
 b1,b2             : byte;
 b                 : byte;
begin
   b:= 0; CodeUsed:= true;
   if (c1 = '-') and (c2 = '-') then begin
      CodeUsed:= false;
      goto skip;
   end;
   if (c1 = '-') then begin
      if IsHex(c2) then begin
         b:= GetHexValue(c2);
      end else begin
         ErrorMsg('Invalid instruction data!');
      end;
   end else begin
      if (c2 = '-') then begin
         if IsHex(c1) then begin
            b:= GetHexValue(c1);
         end else begin
            ErrorMsg('Invalid instruction data!');
         end;
      end else begin
         if (IsHex(c1)) and (IsHex(c2)) then begin
            b1:= GetHexValue(c1);
            b2:= GetHexValue(c2);
            b:= (b1*16) + b2;
         end else begin
            ErrorMsg('Invalid instruction data!');
         end;
      end;
   end;
skip:
   GetCode:= b;
end;

procedure BuildOPCode(c1, c2, c3, c4, r1, r2: char);
label
 skip;
var
 b1,b2             : byte;
 u1,u2             : boolean;
begin
   b1:= GetCode(c1,c2);
   if IsErr then goto skip;
   u1:= CodeUsed;
   b2:= GetCode(c3,c4);
   if IsErr then goto skip;
   u2:= CodeUsed;
   if (not u1) and (not u2) then goto skip;
   if (u1) and (not u2) then begin
      if r1 = '1' then b1:= b1 + GetRegCode(paString[1]);
      if r2 = '1' then b1:= b1 + GetRegCode(paString[2]);
      AddCodeListVALUE(tBYTE, b1);
   end else begin
      if (not u1) and (u2) then begin
         if r1 = '1' then b2:= b2 + GetRegCode(paString[1]);
         if r2 = '1' then b2:= b2 + GetRegCode(paString[2]);
         AddCodeListVALUE(tBYTE, b2);
      end else begin
         if r1 = '1' then b2:= b2 + GetRegCode(paString[1]);
         if r2 = '1' then b2:= b2 + GetRegCode(paString[2]);
         AddCodeListVALUE(tBYTE, b1);
         AddCodeListVALUE(tBYTE, b2);
      end;
   end;
skip:
end;

procedure RunInstr(iinstr: string);
label
 endBuild;
var
 i,id              : longint;
 iEXE              : boolean;
 p1,p2             : word;
 idat              : string;
 v                 : TDefValue;
 b                 : byte;
 u32, mu32         : boolean;
begin
   runInstr2:= false;
   if InstrCount > 0 then begin
      id:= NotFound;
      for i:= 0 to InstrCount - 1 do begin
         if InstrList[i].Name = iinstr then begin
            id:= i; break;
         end;
      end;
      paValue[1]:= 0;
      paString[1]:= '';
      paType[1]:= 0;
      paSpecial[1]:= 0;
      paOpr[1]:= 0;
      paValue[2]:= 0;
      paString[2]:= '';
      paType[2]:= 0;
      paSpecial[2]:= 0;
      paOpr[2]:= 0;
      if id <> NotFound then begin
         GrabParameter(1);
         if not IsErr then begin
            if (paType[1] <> pNone) and (paType[1] <> pSegOffset) then begin
               SkipBlank;
               if not IsENDLine then begin
                  if GetSrcChar = ',' then begin
                     NextChar;
                     SkipBlank;
                     GrabParameter(2);
                  end;
               end;
            end else begin
               if paType[1] <> pSegOffset then begin
                  paValue[2]:= 0;
                  paString[2]:= '';
                  paType[2]:= pNONE;
                  paSpecial[2]:= 0;
                  paOpr[2]:= oprNONE;
               end;
            end;
            if not IsErr then begin
               iEXE:= false;
               for i:= id to InstrCount - 1 do begin
                  if InstrList[i].Name = iinstr then begin
                     p1:= InstrList[i].par1;
                     p2:= InstrList[i].par2;
                     if ((p1 = paType[1]) or (p1 = paSpecial[1])) and
                      ((p2 = paType[2]) or (p2 = paSpecial[2]))
                      then begin
                        iEXE:= true;
                        idat:= InstrList[i].iData;
                        if IsHex(idat[3]) then begin
                           v:= GetHexValue(idat[3]);
                           if CurProcessor = -1 then begin
                              ErrorMsg('No Processor selected!');
                              goto endBuild;
                           end;
                           if ProcessorList[CurProcessor].Value >= v then begin
                              if Segment[getSeg].SegBits = 32 then
                               u32:= true
                              else
                               u32:= false;
                              mu32:= u32;
                              buildCode(idat[1],idat[2]);
                              if IsErr then goto endBuild;
                              if idat[6] = '1' then begin
                                 if not u32 then begin
                                    AddCodeListVALUE(tBYTE, $67);
                                    mu32:= true;
                                 end;
                              end;
                              if idat[7] = '1' then begin
                                 if not u32 then begin
                                    AddCodeListVALUE(tBYTE, $66);
                                 end;
                              end;
                              if idat[8] = '1' then begin
                                 if u32 then begin
                                    AddCodeListVALUE(tBYTE, $67);
                                    mu32:= false;
                                 end;
                              end;
                              if idat[9] = '1' then begin
                                 if u32 then begin
                                    AddCodeListVALUE(tBYTE, $66);
                                 end;
                              end;
                              buildCode(idat[10],idat[11]);
                              if IsErr then goto endBuild;

                              buildOPCode(idat[12],idat[13],idat[14],idat[15],idat[30],idat[31]);
                              if IsErr then goto endBuild;

                              if idat[16] = '1' then begin
                                 b:= GetModRM2(paString[1], paString[2],mu32);
                                 if IsErr then goto endBuild;
                                 AddCodeListVALUE(tBYTE, b);
                              end;
                              if idat[16] = '2' then begin
                                 b:= GetCode(idat[17],idat[18]);
                                 if IsErr then goto endBuild;
                                 b:= GetModRMDigit2(paString[1],b,mu32);
                                 if IsErr then goto endBuild;
                                 AddCodeListVALUE(tBYTE, b);
                              end;
                              if idat[16] = '3' then begin
                                 b:= GetCode(idat[17],idat[18]);
                                 if IsErr then goto endBuild;
                                 b:= GetModRMDigit2(paString[2],b,mu32);
                                 if IsErr then goto endBuild;
                                 AddCodeListVALUE(tBYTE, b);
                              end;
                              if idat[16] = '4' then begin
                                 b:= GetCode(idat[17],idat[18]);
                                 if IsErr then goto endBuild;
                                 b:= GetSpecialModRM2(GetRegCode(paString[1]),b);
                                 if IsErr then goto endBuild;
                                 AddCodeListVALUE(tBYTE, b);
                              end;
                              if idat[16] = '5' then begin
                                 b:= GetCode(idat[17],idat[18]);
                                 if IsErr then goto endBuild;
                                 b:= GetSpecialModRM2(GetRegCode(paString[2]),b);
                                 if IsErr then goto endBuild;
                                 AddCodeListVALUE(tBYTE, b);
                              end;
                              buildCode(idat[19],idat[20]);
                              if IsErr then goto endBuild;

                              if idat[21] = '1' then begin
                                 AddReloc(Segment[getSeg].Offset);
                              end;

                              if idat[22] = '1' then begin
                                 b:= 1;
                                 if idat[4] = '0' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListREL(tBYTE, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListRELID(tBYTE, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                                 if idat[4] = '1' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListREL(tWORD, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListRELID(tWORD, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                                 if idat[4] = '2' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListREL(tDWORD, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListRELID(tDWORD, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                                 if idat[4] = '3' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListREL(tQWORD, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListRELID(tQWORD, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                                 if idat[4] = '4' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListREL(tFARPTR, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListRELID(tFARPTR, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                                 if idat[4] = '5' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListREL(tTENBYTE, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListRELID(tTENBYTE, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                                 if idat[4] = '6' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListREL(tBYTE, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListRELID(tBYTE, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                              end;
                              if idat[23] = '1' then begin
                                 b:= 2;
                                 if idat[5] = '0' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListREL(tBYTE, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListRELID(tBYTE, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                                 if idat[5] = '1' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListREL(tWORD, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListRELID(tWORD, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                                 if idat[5] = '2' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListREL(tDWORD, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListRELID(tDWORD, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                                 if idat[5] = '3' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListREL(tQWORD, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListRELID(tQWORD, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                                 if idat[5] = '4' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListREL(tFARPTR, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListRELID(tFARPTR, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                                 if idat[5] = '5' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListREL(tTENBYTE, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListRELID(tTENBYTE, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                                 if idat[5] = '6' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListREL(tBYTE, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListRELID(tBYTE, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                              end;
                              if idat[24] = '1' then begin
                                 b:= 1;
                                 if idat[4] = '0' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListMEM(tWORD, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListMEMID(tWORD, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                                 if idat[4] = '1' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListMEM(tWORD, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListMEMID(tWORD, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                                 if idat[4] = '2' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListMEM(tDWORD, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListMEMID(tDWORD, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                                 if idat[4] = '3' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListMEM(tQWORD, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListMEMID(tQWORD, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                                 if idat[4] = '4' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListMEM(tFARPTR, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListMEMID(tFARPTR, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                                 if idat[4] = '5' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListMEM(tTENBYTE, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListMEMID(tTENBYTE, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                                 if idat[4] = '6' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListMEM(tBYTE, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListMEMID(tBYTE, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                              end;
                              if idat[25] = '1' then begin
                                 b:= 2;
                                 if idat[5] = '0' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListMEM(tWORD, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListMEMID(tWORD, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                                 if idat[5] = '1' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListMEM(tWORD, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListMEMID(tWORD, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                                 if idat[5] = '2' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListMEM(tDWORD, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListMEMID(tDWORD, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                                 if idat[5] = '3' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListMEM(tQWORD, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListMEMID(tQWORD, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                                 if idat[5] = '4' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListMEM(tFARPTR, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListMEMID(tFARPTR, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                                 if idat[5] = '5' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListMEM(tTENBYTE, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListMEMID(tTENBYTE, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                                 if idat[5] = '6' then begin
                                    v:= GetIdentIndex(paString[b]);
                                    if v = NotFound then begin
                                       AddCodeListMEM(tBYTE, paString[b], paOpr[b], paValue[b]);
                                    end else begin
                                       AddCodeListMEMID(tBYTE, v, paOpr[b], paValue[b]);
                                    end;
                                 end;
                              end;
                              if idat[26] = '1' then begin
                                 b:= 1;
                                 if idat[4] = '0' then begin
                                    AddCodeListVALUE(tBYTE, paValue[b]);
                                 end;
                                 if idat[4] = '1' then begin
                                    AddCodeListVALUE(tWORD, paValue[b]);
                                 end;
                                 if idat[4] = '2' then begin
                                    AddCodeListVALUE(tDWORD, paValue[b]);
                                 end;
                                 if idat[4] = '3' then begin
                                    AddCodeListVALUE(tQWORD, paValue[b]);
                                 end;
                                 if idat[4] = '4' then begin
                                    AddCodeListVALUE(tFARPTR, paValue[b]);
                                 end;
                                 if idat[4] = '5' then begin
                                    AddCodeListVALUE(tTENBYTE, paValue[b]);
                                 end;
                                 if idat[4] = '6' then begin
                                    AddCodeListVALUE(tBYTE, paValue[b]);
                                 end;
                              end;
                              if idat[27] = '1' then begin
                                 b:= 2;
                                 if idat[5] = '0' then begin
                                    AddCodeListVALUE(tBYTE, paValue[b]);
                                 end;
                                 if idat[5] = '1' then begin
                                    AddCodeListVALUE(tWORD, paValue[b]);
                                 end;
                                 if idat[5] = '2' then begin
                                    AddCodeListVALUE(tDWORD, paValue[b]);
                                 end;
                                 if idat[5] = '3' then begin
                                    AddCodeListVALUE(tQWORD, paValue[b]);
                                 end;
                                 if idat[5] = '4' then begin
                                    AddCodeListVALUE(tFARPTR, paValue[b]);
                                 end;
                                 if idat[5] = '5' then begin
                                    AddCodeListVALUE(tTENBYTE, paValue[b]);
                                 end;
                                 if idat[5] = '6' then begin
                                    AddCodeListVALUE(tBYTE, paValue[b]);
                                 end;
                              end;
                              if idat[28] = '1' then begin
                                 AddReloc(Segment[getSeg].Offset);
                                 if idat[32] <> '0' then
                                  AddCodeListSEGOFFSET(tWORD, GetSegIndex('.CODE'), oprDIV, 16)
                                 else
                                  AddCodeListSEGOFFSET(tWORD, GetSegIndex('.CODE'), oprNone, 0);
                              end;
                              if idat[28] = '2' then begin
                                 AddReloc(Segment[getSeg].Offset);
                                 if idat[32] <> '0' then
                                  AddCodeListSEGOFFSET(tWORD, GetSegIndex('.DATA'), oprDIV, 16)
                                 else
                                  AddCodeListSEGOFFSET(tWORD, GetSegIndex('.DATA'), oprNone, 0);
                              end;
                              if idat[28] = '3' then begin
                                 AddReloc(Segment[getSeg].Offset);
                                 if idat[32] <> '0' then
                                  AddCodeListSEGOFFSET(tWORD, GetSegIndex('.SSEG'), oprDIV, 16)
                                 else
                                  AddCodeListSEGOFFSET(tWORD, GetSegIndex('.SSEG'), oprNone, 0);
                              end;
                              if idat[28] = '4' then begin
                                 AddReloc(Segment[getSeg].Offset);
                                 if idat[32] <> '0' then
                                  AddCodeListSEGOFFSET(tWORD, GetSegIndex('.BSS'), oprDIV, 16)
                                 else
                                  AddCodeListSEGOFFSET(tWORD, GetSegIndex('.BSS'), oprNone, 0);
                              end;
                              if idat[28] = '5' then begin
                                 AddReloc(Segment[getSeg].Offset);
                                 if idat[32] <> '0' then
                                  AddCodeListSEGOFFSET(tWORD, GetSegIndex('.FSEG'), oprDIV, 16)
                                 else
                                  AddCodeListSEGOFFSET(tWORD, GetSegIndex('.FSEG'), oprNone, 0);
                              end;
                              if idat[28] = '6' then begin
                                 AddReloc(Segment[getSeg].Offset);
                                 if idat[32] <> '0' then
                                  AddCodeListSEGOFFSET(tWORD, GetSegIndex('.GSEG'), oprDIV, 16)
                                 else
                                  AddCodeListSEGOFFSET(tWORD, GetSegIndex('.GSEG'), oprNone, 0);
                              end;
                              if idat[29] = '1' then begin
                                 AddReloc(Segment[getSeg].Offset);
                                 if idat[32] <> '0' then
                                  AddCodeListSEGOFFSET(tDWORD, GetSegIndex('.CODE'), oprDIV, 16)
                                 else
                                  AddCodeListSEGOFFSET(tDWORD, GetSegIndex('.CODE'), oprNone, 0);
                              end;
                              if idat[29] = '2' then begin
                                 AddReloc(Segment[getSeg].Offset);
                                 if idat[32] <> '0' then
                                  AddCodeListSEGOFFSET(tDWORD, GetSegIndex('.DATA'), oprDIV, 16)
                                 else
                                  AddCodeListSEGOFFSET(tDWORD, GetSegIndex('.DATA'), oprNone, 0);
                              end;
                              if idat[29] = '3' then begin
                                 AddReloc(Segment[getSeg].Offset);
                                 if idat[32] <> '0' then
                                  AddCodeListSEGOFFSET(tDWORD, GetSegIndex('.SSEG'), oprDIV, 16)
                                 else
                                  AddCodeListSEGOFFSET(tDWORD, GetSegIndex('.SSEG'), oprNone, 0);
                              end;
                              if idat[29] = '4' then begin
                                 AddReloc(Segment[getSeg].Offset);
                                 if idat[32] <> '0' then
                                  AddCodeListSEGOFFSET(tDWORD, GetSegIndex('.BSS'), oprDIV, 16)
                                 else
                                  AddCodeListSEGOFFSET(tDWORD, GetSegIndex('.BSS'), oprNone, 0);
                              end;
                              if idat[29] = '5' then begin
                                 AddReloc(Segment[getSeg].Offset);
                                 if idat[32] <> '0' then
                                  AddCodeListSEGOFFSET(tDWORD, GetSegIndex('.FSEG'), oprDIV, 16)
                                 else
                                  AddCodeListSEGOFFSET(tDWORD, GetSegIndex('.FSEG'), oprNone, 0);
                              end;
                              if idat[29] = '6' then begin
                                 AddReloc(Segment[getSeg].Offset);
                                 if idat[32] <> '0' then
                                  AddCodeListSEGOFFSET(tDWORD, GetSegIndex('.GSEG'), oprDIV, 16)
                                 else
                                  AddCodeListSEGOFFSET(tDWORD, GetSegIndex('.GSEG'), oprNone, 0);
                              end;
                              runInstr2:= true;
endBuild:
                              break;
                           end else begin
                              ErrorMsg('This instruction is only available in '
                               + IntToStr(v) + '86-mode and higher!');
                           end;
                        end else begin
                           ErrorMsg('Invalid processor code!');
                        end;
                     end;
                  end;
               end;
               if not iEXE then begin
                  ErrorMsg('Illegal use of [' + iinstr + ']!');
               end;
            end;
         end;
      end;
   end;
end;

end.
